
/*******************************************************************************
*
* File check15.c
*
* Copyright (C) 2005, 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Check of the acceptance step
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "random.h"
#include "start.h"
#include "misc.h"
#include "block.h"
#include "linalg.h"
#include "dirac.h"
#include "sw_term.h"
#include "hmc.h"
#include "global.h"

static const su3_dble ud0={{0.0}};


static void init_ud(void)
{
   su3_alg_dble X;
   su3_dble *u,*um;

   u=pud[VOLUME/2][0];
   um=u+4*VOLUME;

   for (;u<um;u++)
   {
      *u=ud0;
      (*u).c11.re=1.0;
      (*u).c22.re=1.0;
      (*u).c33.re=1.0;      
      random_alg(1,&X);
      ExpX(0.1,&X,u);      
   }

   set_flags(NEW_UD);
   renormalize_ud();
}


static void rot_ud(int n,double t)
{
   int nb,isw,*lk,*lkm;
   su3_dble *ub;
   su3_alg_dble *mom;
   block_t *b;
   hmc_t *hflds;

   b=blk_list(HMC_BLOCKS,&nb,&isw)+n;
   hflds=hmcflds()+n;
   
   ub=(*b).ud[0];
   mom=(*hflds).mom;
   lk=(*hflds).lks;
   lkm=lk+(*hflds).nlks;
   
   for (;lk<lkm;lk++)
   {
      ExpX(t,mom,ub+(*lk));
      mom+=1;
   }

   set_blk_flags(HMC_BLOCKS,NEW_ACTIVE_UD,n);
}


static void hmc_action(hmc_act_t act)
{
   int n,nb,isw;

   blk_list(HMC_BLOCKS,&nb,&isw);

   for (n=0;n<nb;n++)
   {
      act.act0[n]=hmc_acg(n);
      act.act1[n]=0.0;
      act.act2[n]=0.0;
   }
}


static double sum_action(hmc_act_t act)
{
   int nb,n;
   double *a0,*a1,*a2,sm,sall;

   nb=act.nb;
   a0=act.act0;
   a1=act.act1;
   a2=act.act2;   

   sm=0.0;

   for (n=0;n<nb;n++)
      sm+=(a0[n]+a1[n]+a2[n]);

   if (NPROC>1)
   {
      MPI_Reduce(&sm,&sall,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(&sall,1,MPI_DOUBLE,0,MPI_COMM_WORLD);     
   }
   else
      sall=sm;

   return sall;
}


static void random_action(hmc_act_t act)
{
   int nb,n;
   double vol,*a0,*a1,*a2;

   nb=act.nb;
   a0=act.act0;
   a1=act.act1;
   a2=act.act2;
   
   ranlxd(a0,nb);
   ranlxd(a1,nb);
   ranlxd(a2,nb);   
   
   vol=10.0*(double)(VOLUME)/(double)(nb);   

   for (n=0;n<nb;n++)
   {
      a0[n]*=vol;
      a1[n]*=vol;
      a2[n]*=vol;
   }
}


static double diff_action(hmc_act_t act0,hmc_act_t act1)
{
   int nb,n;
   double a0,a1,a2,sm,sall;

   nb=act0.nb;
   a0=0.0;
   a1=0.0;
   a2=0.0;

   for (n=0;n<nb;n++)
   {
      a0+=(act1.act0[n]-act0.act0[n]);
      a1+=(act1.act1[n]-act0.act1[n]);
      a2+=(act1.act2[n]-act0.act2[n]);
   }

   sm=a0+a1+a2;

   if (NPROC>1)
   {
      MPI_Reduce(&sm,&sall,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(&sall,1,MPI_DOUBLE,0,MPI_COMM_WORLD);     
   }
   else
      sall=sm;
   
   return sall;
}


int main(int argc,char *argv[])
{
   int my_rank,n,nb,isw;
   int bc,bs_hmc[4],Ntry,ib,iac,dmy;
   double sact0,sact1,dact,dmax,dH,Pacc,p;
   hmc_t *hflds;
   hmc_act_t act0,act1,act2;
   FILE *flog=NULL,*fin=NULL;

   MPI_Init(&argc,&argv);
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   if (my_rank==0)
   {
      flog=freopen("check15.log","w",stdout);
      fin=freopen("check15.in","r",stdin);
      
      printf("\n");
      printf("Check of the acceptance step\n");
      printf("----------------------------\n\n");

      printf("%dx%dx%dx%d lattice, ",NPROC0*L0,NPROC1*L1,NPROC2*L2,NPROC3*L3);
      printf("%dx%dx%dx%d process grid, ",NPROC0,NPROC1,NPROC2,NPROC3);
      printf("%dx%dx%dx%d local lattice\n\n",L0,L1,L2,L3);

      read_line("bc","%d\n",&bc);
      read_line("bs_hmc","%d %d %d %d\n",
                &bs_hmc[0],&bs_hmc[1],&bs_hmc[2],&bs_hmc[3]);
      read_line("dH","%lf\n",&dH);
      read_line("Ntry","%d",&Ntry);      
      fclose(fin);

      printf("bc = %d\n",bc);
      printf("bs_hmc = %d %d %d %d\n",bs_hmc[0],bs_hmc[1],bs_hmc[2],bs_hmc[3]);
      printf("dH = %.2e\n",dH);
      printf("Number of trials = %d\n\n",Ntry);
   }

   MPI_Bcast(&bc,1,MPI_INT,0,MPI_COMM_WORLD);   
   MPI_Bcast(bs_hmc,4,MPI_INT,0,MPI_COMM_WORLD);
   MPI_Bcast(&dH,1,MPI_DOUBLE,0,MPI_COMM_WORLD);      
   MPI_Bcast(&Ntry,1,MPI_INT,0,MPI_COMM_WORLD);   
   
   start_ranlux(0,1234);
   geometry();
   alloc_ud();

   set_hmc_parms(bs_hmc,bc,1,1,1,0,0,0,0,1,1,1,1,1,0.5,1.0,1.0,1.0,1.0,1.0);

   alloc_hmc_action(&act0);
   alloc_hmc_action(&act1);
   alloc_hmc_action(&act2);
   
   blk_list(HMC_BLOCKS,&nb,&isw);
   Pacc=0.0;
   dmax=0.0;

   for (n=0;n<Ntry;n++)
   {
      init_ud();
      if (bc==-1)
         flipbcd();
      assign_ud2udbgr(HMC_BLOCKS);

      hflds=hmcflds();
   
      for (ib=0;ib<nb;ib++)
      {
         random_alg((*hflds).nlks,(*hflds).mom);
         rot_ud(ib,0.1);
         hflds+=1;
      }

      hmc_action(act1);
      hmc_action(act2);
      act2.act0[0]+=(dH/(double)(NPROC));
      swap_active(HMC_BLOCKS);
      hmc_action(act0);

      iac=accept_step(act1,act2,&dact);
      Pacc+=(double)(iac);

      dact=fabs(1.0-dact/dH);
      if (dact>dmax)
         dmax=dact;
      
      if (NPROC>1)
      {
         dmy=iac;
         MPI_Bcast(&dmy,1,MPI_INT,0,MPI_COMM_WORLD);

         error(dmy!=iac,1,"main [check15.c]","iac is not globally the same");
      }

      hmc_action(act2);
      
      if (iac==0)
         error(diff_action(act0,act2)!=0.0,1,"main [check15.c]",
               "Fields swapping is incorrect");
      else
         error(diff_action(act1,act2)!=0.0,1,"main [check15.c]",
               "Fields swapping is incorrect");

      error(query_flags(BCD_FLIPPED)!=0,1,"main [check15.c]",
            "Boundary links are flipped");
   }

   Pacc/=(double)(Ntry);

   if (dH>0.0)
      p=exp(-dH);
   else
      p=1.0;
   
   error_chk();
 
   if (my_rank==0)
   {
      printf("No errors discovered\n");
      printf("Acceptance rate = %.3f ",Pacc);
      printf("(expect %.3f+-%.3f)\n",p,sqrt(p*(1.0-p)/(double)(Ntry)));
      printf("Deviation of action difference (statistical test) = %.1e\n",dmax);
   }      

   if (bc==-1)
      flipbcd();

   random_action(act0);
   random_action(act1);
   accept_step(act0,act1,&dact);

   sact0=sum_action(act0);
   sact1=sum_action(act1);
   dmax=fabs(1.0-(sact1-sact0)/dact);

   if (my_rank==0)
   {
      printf("Deviation of action difference (random action) = %.1e\n\n",dmax);
      fclose(flog);   
   }      
   
   MPI_Finalize();   
   exit(0);
}
