
/*******************************************************************************
*
* File read3a.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Reads and evaluates data from the *.dat files created by the program run3.
* The file to be read has to be specified on the command line
*
* This program prints the averages of the plaquette, the factor exp(-dH),
* the acceptance rate and the average solver iteration numbers to stdout
*
*******************************************************************************/

#define MAIN_PROGRAM

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "start.h"
#include "misc.h"
#include "extras.h"

typedef struct
{
   int nt,ns,iac;
   int stat_rop,stat_frf,stat_acf;
   int stat_gcr_frf[3],stat_gcr_acf[3];
   int stat_dfl_mds,stat_dfl_upd[2];
   double dact,avpl;
} dat_t;

static int nms,nfirst,nlast,neff;
static dat_t *adat;


static int read_dat(int n,dat_t *ndat,FILE *fin)
{
   int i,iw,ic,ie;
   stdint_t istd[15];
   double dstd[2];

   ic=0;
   ie=endianness();
   
   for (i=0;i<n;i++)
   {
      iw =fread(istd,sizeof(stdint_t),15,fin);
      iw+=fread(dstd,sizeof(double),2,fin);

      if (iw==17)
         ic+=1;
      else
         break;

      if (ie==BIG_ENDIAN)
      {
         bswap_int(15,istd);
         bswap_double(2,dstd);
      }
      
      (*ndat).nt=(int)(istd[0]);
      (*ndat).ns=(int)(istd[1]);
      (*ndat).iac=(int)(istd[2]);

      (*ndat).stat_rop=(int)(istd[3]);
      (*ndat).stat_frf=(int)(istd[4]);
      (*ndat).stat_acf=(int)(istd[5]);       

      (*ndat).stat_gcr_frf[0]=(int)(istd[ 6]);
      (*ndat).stat_gcr_frf[1]=(int)(istd[ 7]);
      (*ndat).stat_gcr_frf[2]=(int)(istd[ 8]);      

      (*ndat).stat_gcr_acf[0]=(int)(istd[ 9]);
      (*ndat).stat_gcr_acf[1]=(int)(istd[10]);
      (*ndat).stat_gcr_acf[2]=(int)(istd[11]);       

      (*ndat).stat_dfl_mds=(int)(istd[12]);
      (*ndat).stat_dfl_upd[0]=(int)(istd[13]);
      (*ndat).stat_dfl_upd[1]=(int)(istd[14]);      
      
      (*ndat).dact=dstd[0];
      (*ndat).avpl=dstd[1];

      ndat+=1;
   }
   
   return ic;
} 


static void read_file(char *file)
{
   int n,ir;
   long ipos;
   dat_t ndat;
   FILE *fdat;

   fdat=fopen(file,"rb");
   error(fdat==NULL,1,"read_file [read3a.c]","Unable to open data file");

   printf("Read data from file %s\n\n",file);
   ipos=ftell(fdat);
   ir=0;
      
   for (n=0;feof(fdat)==0;n++)
      ir+=read_dat(1,&ndat,fdat);

   error((ir!=(n-1))||(ferror(fdat)!=0),1,"read_file [read3a.c]",
         "Incorrect read count or read error");   
   
   nms=n-1;
   adat=amalloc(nms*sizeof(*adat),3);
   error(adat==NULL,1,"read_file [read3a.c]",
         "Unable to allocate data array");

   fseek(fdat,ipos,SEEK_SET);
   ir=read_dat(nms,adat,fdat);

   error((ir!=nms)||(feof(fdat)!=0)||(ferror(fdat)!=0),1,"read_file [read3a.c]",
         "Error while reading data file");
   fclose(fdat);
}


static void select_range(void)
{
   int n,nf,nl,nt,np,dn,ie;
   
   printf("There are %d measurements (trajectories no %d - %d)\n",
          nms,adat[0].nt,adat[nms-1].nt);
   printf("Range [nfirst,nlast] of trajectories to analyse: ");
   scanf("%d %d",&nfirst,&nlast);

   nf=0;
   nl=0;
      
   for (n=0;n<nms;n++)
   {
      nt=adat[n].nt;
      
      if (nt<nfirst)
         nf+=1;

      if (nt<=nlast)
         nl+=1;
   }

   nfirst=nf;
   nlast=nl;
   neff=nlast-nfirst;
      
   printf("Keep %d measurements (trajectories no %d - %d)\n\n",
          neff,adat[nfirst].nt,adat[nlast-1].nt);

   if (neff>1)
   {
      np=adat[nfirst].nt;
      dn=adat[nfirst+1].nt-adat[nfirst].nt;

      if (dn<=0)
         ie=1;
      else
      {
         ie=0;
         
         for (n=(nfirst+1);n<nlast;n++)
         {
            nt=adat[n].nt;

            if ((nt-np)!=dn)
            {
               ie=2;
               break;
            }

            np=nt;
         }
      }

      error(ie!=0,1,"select_range [read3a.c]",
            "Irregular trajectory numbers between successive\n"
            "measurements in selected range");
   }
}


static double f(int nx,double x[])
{
   return x[0];
}


int main(int argc,char *argv[])
{
   int n;
   double **pa,*a,abar;
   
   error(argc!=2,1,"main [read3a.c]","Syntax: read3a <filename>");

   printf("\n");
   printf("DD-HMC simulation of two-flavour O(a) improved QCD\n");
   printf("--------------------------------------------------\n\n");

   read_file(argv[1]);
   select_range();

   pa=amalloc(1*sizeof(*pa),3);
   a=amalloc(neff*sizeof(double),3);
   error((pa==NULL)||(a==NULL),1,"main [read3a.c]",
         "Unable to allocate data array");
   *pa=a;
      
   for (n=0;n<neff;n++)
      a[n]=adat[nfirst+n].avpl;

   printf("In the following, the integrated autocorrelations and the associated"
          "\nstatistical error sigma of each quantity is estimated ");

   if (neff>100)
      printf("using the\nnumerically determined "
             "autocorrelation function.\n\n");
   else
      printf("by binning the\ndata and by calculating "
             "the jackknife errors of the binned series.\n\n");

   printf("The autocorrelation times are given in numbers of measurements\n"
          "separated by %d trajectories.\n\n",
          adat[nfirst+1].nt-adat[nfirst].nt);

   printf("Plaquette\n"
          "---------\n\n");
   
   if (neff>=100)
      abar=print_auto(neff,a);
   else
      abar=print_jack(1,neff,pa,f);
   printf("<tr{U(p)}> = %1.6f\n\n",abar);

   printf("Expectation value of exp(-dH)\n"
          "-----------------------------\n\n");
   
   for (n=0;n<neff;n++)
      a[n]=exp(-adat[nfirst+n].dact);

   if (neff>=100)
      abar=print_auto(neff,a);
   else
      abar=print_jack(1,neff,pa,f);
   printf("<exp(-dH)> = %1.6f\n\n",abar);

   printf("Acceptance rate\n"
          "---------------\n\n");
   
   for (n=0;n<neff;n++)
      a[n]=(double)(adat[nfirst+n].iac);

   if (neff>=100)
      abar=print_auto(neff,a);
   else
      abar=print_jack(1,neff,pa,f);   
   printf("<iac> = %1.2f\n\n",abar);

   printf("Average solver iteration numbers\n"
          "--------------------------------\n\n");
   
   for (n=0;n<neff;n++)
      a[n]=(double)(adat[nfirst+n].stat_gcr_frf[0]);

   if (neff>=100)
      abar=print_auto(neff,a);
   else
      abar=print_jack(1,neff,pa,f);     
   printf("<stat_gcr_frf[0]> = %1.2f\n\n",abar);

   for (n=0;n<neff;n++)
      a[n]=(double)(adat[nfirst+n].stat_gcr_frf[1]);

   if (neff>=100)
      abar=print_auto(neff,a);
   else
      abar=print_jack(1,neff,pa,f);     
   printf("<stat_gcr_frf[1]> = %1.2f\n\n",abar);   

   for (n=0;n<neff;n++)
      a[n]=(double)(adat[nfirst+n].stat_gcr_frf[2]);

   if (neff>=100)
      abar=print_auto(neff,a);
   else
      abar=print_jack(1,neff,pa,f);     
   printf("<stat_gcr_frf[2]> = %1.2f\n\n",abar);   

   exit(0);
}
