
/*******************************************************************************
*
* File Zops.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of the little Dirac operator
*
* The externally accessible functions are
*
*   zop_dble_t *zopnohat_dble(void)
*     Returns the pointer to a structure containing the double-precision
*     matrix arrays that define the little Dirac operator (see the notes
*     below)
*
*   zop_dble_t *zophat_dble(void)
*     Returns the pointer to a structure containing the double-precision
*     matrix arrays that define the even-odd preconditioned form of the
*     little Dirac operator. The matrices are computed starting from those
*     returned by zopnohat_dble()
*
*   zop_t *zopnohat(void)
*     Returns the pointer to a structure containing the single-precision
*     matrix arrays that define the little Dirac operator. The matrices
*     are computed starting from those returned by zopnohat_dble()
*
*   zop_t *zophat(void)
*     Returns the pointer to a structure containing the single-precision
*     matrix arrays that define the even-odd preconditioned form of the
*     little Dirac operator. The matrices are computed starting from those
*     returned by zophat_dble()
*
*   void dfl_little_modes(complex **v,complex **zv,complex_dble **vd,
*                         complex_dble **zvd,complex_dble *a)
*     Sets the first Ns elements of v[] and vd[] to the addresses of Ns
*     internally stored, orthonormal single- and double-precision vector
*     fields that are used to deflate the even-odd preconditioned little
*     Dirac operator. The addresses of Ns other fields equal to Zhat*v[]
*     and Zhat*vd[] are assigned to zv[] and zvd[]. On exit the elements
*     of the *inverse* of the Ns x Ns matrix (vd[i],zvd[j]) are assigned
*     to a[i*Ns+j]
*
*   void free_zops(void)
*     Frees the memory space occupied by the static matrix arrays that
*     define the little Dirac operator and its even-odd preconditioned
*     form
*
* Notes:
*
* The deflation subspace is defined by the fields (*b).sd[1],..,(*b).sd[Ns]
* on the blocks b of the DFL_BLOCKS grid. The restriction of the (improved)
* Wilson-Dirac operator Dw to the deflation subspace is referred to as the
* little Dirac operator. It is completely specified by the matrix elements
*
*   (v_{i,j},Dw*v_{k,l})
*
*   v_{i,0},v_{i,1},..,v_{i,Ns-1} = basis vectors on the i'th block
*
* The program zopnohat_dble() returns a pointer to a zop_dble_t structure
* in which the matrix elements are stored in complex arrays according to 
*
*  zee: (v_{i,*},Dw*v_{i,*}),  i=even blocks
*
*  zoo: (v_{i,*},Dw*v_{i,*}),  i=odd blocks
*
*  zoe: (v_{i,*},Dw*v_{j,*}),  i=odd blocks, j=inn[i][ifc], ifc=0,..,7
*
*  zeo: (v_{j,*},Dw*v_{i,*}),  i=odd blocks, j=inn[i][ifc], ifc=0,..,7
*
* (the even-odd ordering of the blocks and the related notation is defined
* at the beginning of the module vgrid.c) 
*
* In the case of the even-odd preconditioned little Dirac operator, the
* arrays contain zee^{-1}, zoo^{-1}, zoo^{-1}*zoe and zee^{-1}*zeo, where
* zee,zoo,zoe and zeo are the matrices computed by zopnohat_dble()
* 
* The programs in this module perform global operations and must be called
* simultaneously on all processes. The deflation subspace must be properly 
* set up, using the programs dfl_subspace() or dfl_modes(), before the little
* Dirac operator can be computed (the programs check this by querying the 
* flags data base). Apart from this, no further preparations are required 
*
* Once the little Dirac operator is calculated, its matrix elements are
* stored in static arrays that will only be updated when needed, i.e. when
* the lattice parameters, the global double-precision gauge field or the
* deflation subspace have changed 
*
*******************************************************************************/

#define ZOPS_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "flags.h"
#include "misc.h"
#include "block.h"
#include "linalg.h"
#include "sw_term.h"
#include "dirac.h"
#include "dfl.h"
#include "global.h"

typedef struct
{
   int vol,volh;
   int *ippl,*ippr;
   su3_dble *ud;
   spinor_dble **sdl,**sdr,**buf;
} bbs_t;

static int Nss=0,nvhs;
static int bsv[4]={0,0,0,0};
static int *ipp,nbbe[8],nbbo[8],ofsbbe[8],ofsbbo[8],(*inn)[8];
static complex **vs,**zvs,**vm;
static complex_dble **vds,**zvds,*cs1,*cs2,*ads;
static complex_dble *sndbuf[8],*rcvbuf[8],*sndhop[4],*rcvhop[4];
static bbs_t bbs[4];
static zop_dble_t zop0={0,0,0,NULL,NULL,NULL,NULL};
static zop_dble_t zop1={0,0,0,NULL,NULL,NULL,NULL};
static zop_t zop2={0,0,0,NULL,NULL,NULL,NULL};
static zop_t zop3={0,0,0,NULL,NULL,NULL,NULL};


static int ibsw(int ib,int nb,int isw)
{
   if (isw!=0)
   {
      ib-=(nb/2);
      
      if (ib<0)
         ib+=nb;
   }

   return ib;
}


static int ibnn(int ib,int ifc,int nb,int isw,int *ibn)
{
   int is,in;

   is=ibsw(ib,nb,isw);
   in=inn[is][ifc];

   if (in>=nb)
   {
      (*ibn)=1;
      ifc=ifc^0x1;

      for (;is<nb;)
      {
         in=is;
         is=inn[is][ifc];
      }
   }
   else
      (*ibn)=0;

   return ibsw(in,nb,isw);
}


static void set_bbs(void)
{
   int nb,isw;
   int mu,vol,bvol,(*idn)[4];
   int i,ix,iy;
   int *ip0,*ip1,*ip2;
   block_t *b;
   bndry_t *bb;

   b=blk_list(DFL_BLOCKS,&nb,&isw);
   bb=(*b).bb+1;
   bvol=(*b).vol;
   idn=(*b).idn;

   for (mu=0;mu<4;mu++)
   {
      vol=(*bb).vol;
      
      bbs[mu].vol=vol;
      bbs[mu].volh=vol/2;

      ip0=(*bb).ipp;
      ip1=bbs[mu].ippl;
      ip2=bbs[mu].ippr;
      
      for (i=0;i<vol;i++)
      {
         ix=(*ip0);
         (*ip1)=ix;
         iy=ix;

         for (;iy<bvol;)
         {
            ix=iy;
            iy=idn[iy][mu];
         }

         (*ip2)=ix;
         ip0+=1;
         ip1+=1;
         ip2+=1;
      }

      bb+=2;
   }
}


static void alloc_bbs(void)
{
   int nb,isw,Ns;
   int mu,vol,vmax,vtot,i;
   int *ip0;
   spinor_dble **sd,*sd0,*rs;
   block_t *b;
   bndry_t *bb;

   b=blk_list(DFL_BLOCKS,&nb,&isw);
   bb=(*b).bb+1;
   vmax=0;
   vtot=0;

   for (mu=0;mu<4;mu++)
   {
      vol=(*bb).vol;
      
      if (vmax<vol)
         vmax=vol;

      vtot+=vol;
      bb+=2;
   }

   Ns=Nss;
   ip0=amalloc(2*vtot*sizeof(*ip0),3);
   sd=amalloc(12*Ns*sizeof(*sd),3);
   sd0=amalloc(3*Ns*vmax*sizeof(*sd0),ALIGN);   
   error((ip0==NULL)||(sd==NULL)||(sd0==NULL),1,"alloc_bbs [Zops.c]",
         "Unable to allocate auxiliary arrays");

   bb=(*b).bb+1;
   
   for (mu=0;mu<4;mu++)
   {
      vol=(*bb).vol;
      
      bbs[mu].ippl=ip0;
      ip0+=vol;
      bbs[mu].ippr=ip0;
      ip0+=vol;

      bbs[mu].ud=(*bb).ud;

      bbs[mu].sdl=sd;
      sd+=Ns;
      bbs[mu].sdr=sd;
      sd+=Ns;      
      bbs[mu].buf=sd;
      sd+=Ns;

      rs=sd0;
      
      for (i=0;i<Ns;i++)
      {
         bbs[mu].sdl[i]=rs;
         rs+=vol;
      }

      for (i=0;i<Ns;i++)
      {
         bbs[mu].sdr[i]=rs;
         rs+=vol;
      }

      for (i=0;i<Ns;i++)
      {
         bbs[mu].buf[i]=rs;
         rs+=vol;
      }      
      
      bb+=2;
   }

   set_bbs();
}


static void alloc_modes(void)
{
   int nb,nbb,i;
   int Ns,nvh,nvec;
   complex *v;
   complex_dble *vd;

   Ns=Nss;
   nb=zop0.nb;
   nbb=zop0.nbb;
   nvh=(nb/2)*Ns;
   nvec=(nb+nbb/2)*Ns;

   vs=amalloc(3*Ns*sizeof(*vs),3);
   vds=amalloc(2*Ns*sizeof(*vds),3);
   v=amalloc(2*Ns*nvec*sizeof(*v),ALIGN);
   vd=amalloc((2*Ns*nvec+3*Ns*Ns)*sizeof(*vd),ALIGN);   
   
   error((vs==NULL)||(vds==NULL)||(v==NULL)||(vd==NULL),1,
         "alloc_modes [Zops.c]","Unable to allocate vector fields");

   zvs=vs+Ns;
   vm=zvs+Ns;
   zvds=vds+Ns;
   
   for (i=0;i<Ns;i++)
   {
      vs[i]=v;
      v+=nvec;
      vds[i]=vd;
      vd+=nvec;
   }

   for (i=0;i<Ns;i++)
   {
      zvs[i]=v;
      v+=nvec;
      zvds[i]=vd;
      vd+=nvec;
   }

   cs1=vd;
   cs2=cs1+Ns*Ns;
   ads=cs2+Ns*Ns;
   nvhs=nvh;
}


static void alloc_zop0(void)
{
   int nb,isw,nbb,nbe,*bs;
   int n,ifc,Ns,nmat;
   complex_dble *w;
   dfl_parms_t dfl;
   block_t *b;

   dfl=dfl_parms();   
   b=blk_list(DFL_BLOCKS,&nb,&isw);
   Ns=dfl.Ns;
   bs=(*b).bs;
   
   if ((Ns==Nss)&&(bs[0]==bsv[0])&&(bs[1]==bsv[1])&&(bs[2]==bsv[2])&&
       (bs[3]==bsv[3]))
      return;

   if (Nss>0)
   {
      afree(bbs[0].sdl[0]);
      afree(bbs[0].sdl);
      afree(bbs[0].ippl);

      afree(vs[0]);
      afree(vds[0]);
      afree(vs);
      afree(vds);
   }   
   
   Nss=Ns;   
   bsv[0]=bs[0];
   bsv[1]=bs[1];
   bsv[2]=bs[2];
   bsv[3]=bs[3];

   ipp=vipp(nbbe,nbbo,ofsbbe,ofsbbo);
   inn=vinn(&n);
   nbb=0;
   nbe=0;
   
   for (ifc=0;ifc<8;ifc++)
   {
      nbb+=(nbbe[ifc]+nbbo[ifc]);

      if ((ifc&0x1)==0)
         nbe+=(2*nbbe[ifc]);
      else
         nbe+=(2*nbbo[ifc]);
   }

   zop0.nb=nb;
   zop0.nbb=nbb;
   zop0.Ns=Ns;
   
   if (zop0.zee!=NULL)
      afree(zop0.zee);

   nmat=Ns*Ns;
   n=(nb/2)*nmat;
   if (nbe<nbb)
      nbe=nbb;
   zop0.zee=amalloc((18*n+nbe*nmat)*sizeof(complex_dble),ALIGN);
   error(zop0.zee==NULL,1,"alloc_zop0 [Zops.c]",
         "Unable to allocate space for the matrix elements");

   zop0.zoo=zop0.zee+n;
   zop0.zoe=zop0.zoo+n;
   zop0.zeo=zop0.zoe+8*n;

   w=zop0.zeo+8*n;

   for (ifc=0;ifc<8;ifc++)
   {
      if (nbbo[ifc]>0)
      {
         sndbuf[ifc]=w;
         w+=nbbo[ifc]*nmat;
      }
   }

   for (ifc=0;ifc<8;ifc++)
   {
      if (nbbe[ifc]>0)
      {
         rcvbuf[ifc]=w;
         w+=nbbe[ifc]*nmat;
      }
   }

   w=zop0.zeo+8*n;
   
   for (ifc=1;ifc<8;ifc+=2)
   {
      if (nbbo[ifc]>0)
      {
         sndhop[ifc/2]=w;
         w+=2*nbbo[ifc]*nmat;
      }
   }

   for (ifc=0;ifc<8;ifc+=2)
   {
      if (nbbe[ifc]>0)
      {
         rcvhop[ifc/2]=w;
         w+=2*nbbe[ifc]*nmat;
      }
   }

   alloc_bbs();
   alloc_modes();
}


static void alloc_zop1(void)
{
   int n;
   
   if ((zop1.nb==zop0.nb)&&(zop1.nbb==zop0.nbb)&&(zop1.Ns==zop0.Ns))
      return;

   zop1.nb=zop0.nb;
   zop1.nbb=zop0.nbb;
   zop1.Ns=zop0.Ns;   
   
   if (zop1.zee!=NULL)
      afree(zop1.zee);

   n=(zop1.nb/2)*zop1.Ns*zop1.Ns;   
   zop1.zee=amalloc(18*n*sizeof(complex_dble),ALIGN);
   error(zop1.zee==NULL,1,"alloc_zop1 [Zops.c]",
         "Unable to allocate space for the matrix elements");

   zop1.zoo=zop1.zee+n;
   zop1.zoe=zop1.zoo+n;
   zop1.zeo=zop1.zoe+8*n;
}


static void alloc_zop2(void)
{
   int n;

   if ((zop2.nb==zop0.nb)&&(zop2.nbb==zop0.nbb)&&(zop2.Ns==zop0.Ns))
      return;

   zop2.nb=zop0.nb;
   zop2.nbb=zop0.nbb;
   zop2.Ns=zop0.Ns; 
   
   if (zop2.zee!=NULL)
      afree(zop2.zee);

   n=(zop2.nb/2)*zop2.Ns*zop2.Ns;   
   zop2.zee=amalloc(18*n*sizeof(complex),ALIGN);
   error(zop2.zee==NULL,1,"alloc_zop2 [Zops.c]",
         "Unable to allocate space for the matrix elements");

   zop2.zoo=zop2.zee+n;
   zop2.zoe=zop2.zoo+n;
   zop2.zeo=zop2.zoe+8*n;
}


static void alloc_zop3(void)
{
   int n;

   if ((zop3.nb==zop1.nb)&&(zop3.nbb==zop1.nbb)&&(zop3.Ns==zop1.Ns))
      return;

   zop3.nb=zop1.nb;
   zop3.nbb=zop1.nbb;
   zop3.Ns=zop1.Ns; 
   
   if (zop3.zee!=NULL)
      afree(zop3.zee);

   n=(zop3.nb/2)*zop3.Ns*zop3.Ns;   
   zop3.zee=amalloc(18*n*sizeof(complex),ALIGN);
   error(zop3.zee==NULL,1,"alloc_zop3 [Zops.c]",
         "Unable to allocate space for the matrix elements");

   zop3.zoo=zop3.zee+n;
   zop3.zoe=zop3.zoo+n;
   zop3.zeo=zop3.zoe+8*n;
}


static void copy_ud2blk(block_t *b)
{
   int *imb,*imm,ifc;
   su3_dble *ub,*u;

   imb=(*b).imb+(*b).volh;
   imm=imb+(*b).volh;
   ub=(*b).ud[0];

   for (;imb<imm;imb++)
   {
      u=pud[*imb][0];

      for (ifc=0;ifc<8;ifc++)
      {
         (*ub)=(*u);
         u+=1;
         ub+=1;
      }
   }
}


static void copy_swd2blk(block_t *b)
{
   int *imb,*imm;
   pauli_dble *pb,*p;

   pb=(*b).swd;
   imb=(*b).imb;
   imm=imb+(*b).vol;

   for (;imb<imm;imb++)
   {
      p=swd+2*(*imb);

      (*pb)=(*p);
      pb+=1;
      (*pb)=(*(p+1));
      pb+=1;
   }
}


static void compute_Zdiag(int ib)
{
   int nb,isw,Ns,vol,i,j;
   complex_dble *zmat;
   spinor_dble **sd;
   block_t *b;

   b=blk_list(DFL_BLOCKS,&nb,&isw)+ib;
   
   copy_ud2blk(b);
   copy_swd2blk(b);

   ib=ibsw(ib,nb,isw);
   Ns=Nss;
   zmat=zop0.zee+ib*Ns*Ns;
   vol=(*b).vol;
   sd=(*b).sd;
      
   for (i=0;i<Ns;i++)
   {
      Qnohat_blk_dble(b,i+1,0);
      mulg5_dble(vol,sd[0]);
         
      for (j=0;j<Ns;j++)
         zmat[j*Ns+i]=spinor_prod_dble(vol,0,sd[j+1],sd[0]);
   }
}


static void extract_ubb(int ib,int mu)
{
   int nb,isw,n;
   int *ip,*ipm,*imb;
   su3_dble *ubb,*u;
   block_t *b;

   b=blk_list(DFL_BLOCKS,&nb,&isw);

   ubb=bbs[mu].ud;   
   imb=b[ib].imb;
   ip=bbs[mu].ippl;
   ipm=ip+bbs[mu].volh;

   for (;ip<ipm;ip++)
   {
      u=pud[imb[*ip]][mu];
      (*ubb)=(*u);
      ubb+=1;
   }
   
   ib=ibnn(ib,2*mu+1,nb,isw,&n);
   imb=b[ib].imb;
   ip=bbs[mu].ippr+bbs[mu].volh;
   ipm=ip+bbs[mu].volh;
   
   for (;ip<ipm;ip++)
   {
      u=pud[imb[*ip]][mu]+1;
      (*ubb)=(*u);
      ubb+=1;
   }
}


static void copy_sbb(int mu)
{
   int nbuf,tag;
   int np,saddr,raddr;
   double *sbuf,*rbuf;
   MPI_Status stat;
   
   np=(cpr[0]+cpr[1]+cpr[2]+cpr[3])&0x1;

   nbuf=24*bbs[mu].vol*Nss;   

   saddr=npr[2*mu];
   raddr=npr[2*mu+1];
   sbuf=(double*)(bbs[mu].buf[0]);
   rbuf=(double*)(bbs[mu].sdr[0]);
   tag=mpi_tag();   
   
   if (np==0)
   {
      MPI_Send(sbuf,nbuf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
      MPI_Recv(rbuf,nbuf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
   }
   else
   {
      MPI_Recv(rbuf,nbuf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
      MPI_Send(sbuf,nbuf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
   }
}


static void extract_sbb(int ib,int mu)
{
   int nb,isw,volh,Ns;
   int *ip,ibn,i;
   su3_dble *ud;
   spinor_dble **bsd,**sd;
   block_t *b;

   b=blk_list(DFL_BLOCKS,&nb,&isw);
   
   Ns=Nss;   
   volh=bbs[mu].volh;
   ip=bbs[mu].ippl;
   ud=bbs[mu].ud;
   bsd=b[ib].sd;
   sd=bbs[mu].sdl;

   for (i=0;i<Ns;i++)
   {
      apply_Udag2sd(volh,ip,ud,bsd[i+1],sd[i]);
      gather_sd(volh,ip+volh,bsd[i+1],sd[i]+volh);
   }

   ib=ibnn(ib,2*mu+1,nb,isw,&ibn);
   ip=bbs[mu].ippr;
   bsd=b[ib].sd;

   if (ibn==0)
      sd=bbs[mu].sdr;
   else
      sd=bbs[mu].buf;
   
   for (i=0;i<Ns;i++)
   {
      gather_sd(volh,ip,bsd[i+1],sd[i]);
      apply_U2sd(volh,ip+volh,ud+volh,bsd[i+1],sd[i]+volh);
   }   

   if (ibn!=0)
     copy_sbb(mu);
}


static void compute_Zhop(int ib,int mu)
{
   int nb,isw,nbh,in;
   int Ns,nmat,n,vol,i,j;
   complex_dble sp[2],*z,*zp;
   spinor_dble **sdl,**sdr;

   extract_ubb(ib,mu);
   extract_sbb(ib,mu);
   
   blk_list(DFL_BLOCKS,&nb,&isw);
   ib=ibsw(ib,nb,isw);
   nbh=nb/2;
   Ns=Nss;
   nmat=Ns*Ns;

   if (ib<nbh)
   {
      in=inn[ib][2*mu+1];

      if (in<nb)
      {
         n=(8*(in-nbh)+2*mu)*nmat;
         z=zop0.zeo+n;
         zp=zop0.zoe+n;
      }
      else
      {
         z=sndhop[mu]+2*(in-nb-ofsbbo[2*mu+1])*nmat;
         zp=z+nmat;
      }
   }
   else
   {
      n=(8*(ib-nbh)+2*mu+1)*nmat;
      z=zop0.zoe+n;
      zp=zop0.zeo+n;
   }
   
   vol=bbs[mu].vol;
   sdl=bbs[mu].sdl;
   sdr=bbs[mu].sdr;
   
   for (i=0;i<Ns;i++)
   {
      for (j=0;j<Ns;j++)
      {
         spinor_prod_gamma[mu](vol,sdl[i],sdr[j],sp);

         z[i*Ns+j].re=(-0.5)*(sp[0].re-sp[1].re);
         z[i*Ns+j].im=(-0.5)*(sp[0].im-sp[1].im);

         zp[j*Ns+i].re=(-0.5)*(sp[0].re+sp[1].re);
         zp[j*Ns+i].im=( 0.5)*(sp[0].im+sp[1].im);         
      }
   }
}


static void copy_zbb(void)
{
   int mu,ifc,nbuf,tag;
   int *ibb,*ibm,nmat,nbh,n;
   int np,saddr,raddr;
   double *sbuf,*rbuf;
   complex_dble *z,*zm,*zmat;
   MPI_Status stat;
   
   np=(cpr[0]+cpr[1]+cpr[2]+cpr[3])&0x1;
   nmat=Nss*Nss;
   nbh=zop0.nb/2;

   for (mu=0;mu<4;mu++)
   {
      ifc=2*mu;
      nbuf=4*nbbo[ifc+1]*nmat;

      if (nbuf>0)
      {
         saddr=npr[ifc+1];
         raddr=npr[ifc];
         sbuf=(double*)(sndhop[mu]);
         rbuf=(double*)(rcvhop[mu]);

         tag=mpi_tag();   
   
         if (np==0)
         {
            MPI_Send(sbuf,nbuf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
            MPI_Recv(rbuf,nbuf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
         }
         else
         {
            MPI_Recv(rbuf,nbuf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
            MPI_Send(sbuf,nbuf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
         }

         ibb=ipp+ofsbbe[ifc];
         ibm=ibb+nbbe[ifc];
         z=rcvhop[mu];

         for (;ibb<ibm;ibb++)
         {
            n=(8*((*ibb)-nbh)+ifc)*nmat;
            
            zmat=zop0.zeo+n;
            zm=z+nmat;

            for (;z<zm;z++)
            {
               (*zmat).re=(*z).re;
               (*zmat).im=(*z).im;
               zmat+=1;
            }

            zmat=zop0.zoe+n;
            zm=z+nmat;

            for (;z<zm;z++)
            {
               (*zmat).re=(*z).re;
               (*zmat).im=(*z).im;
               zmat+=1;
            }
         }         
      }
   }
}


zop_dble_t *zopnohat_dble(void)
{
   int nb,ib,mu;
   
   if (query_flags(ZOPNOHATDBLE_UP2DATE)==1)
      return &zop0;

   error_root(query_flags(DFLMODES_UP2DATE)!=1,1,"zopnohat_dble [Zops.c]",
              "The deflation subspace is not allocated or is not up-to-date");
   
   if ((query_flags(SWD_UP2DATE)!=1)||(query_flags(SWD_E_INVERTED)!=0)||
       (query_flags(SWD_O_INVERTED)!=0))
   {
      sw_term();
   }

   alloc_zop0();   
   nb=zop0.nb;
   
   for (ib=0;ib<nb;ib++)
   {
      compute_Zdiag(ib);
      
      for (mu=0;mu<4;mu++)
         compute_Zhop(ib,mu);
   }

   if (NPROC>1)
      copy_zbb();

   set_flags(COMPUTED_ZOPNOHATDBLE);

   return &zop0;
}


static void copy_zee(void)
{
   int np,Ns,nmat;
   int ifc,ifs,ifr,tag,saddr,raddr,nbuf;
   int *ibb,*ibm;
   double *sbuf,*rbuf;
   complex_dble *zmat,*z,*zm;
   MPI_Status stat;

   Ns=zop1.Ns;
   nmat=Ns*Ns;
   np=(cpr[0]+cpr[1]+cpr[2]+cpr[3])&0x1;
   
   for (ifc=0;ifc<8;ifc++)
   {
      if (((cpr[ifc/2]&0x1)^np)==0)
      {
         ifs=ifc;
         ifr=ifc^0x1;
      }
      else
      {
         ifs=ifc^0x1;
         ifr=ifc;
      }

      nbuf=2*nbbo[ifs]*nmat;
      tag=mpi_tag();

      if (nbuf>0)
      {
         ibb=ipp+ofsbbo[ifs];
         ibm=ibb+nbbo[ifs];
         z=sndbuf[ifs];

         for (;ibb<ibm;ibb++)
         {
            zmat=zop1.zee+(*ibb)*nmat;
            zm=z+nmat;

            for (;z<zm;z++)
            {
               (*z).re=(*zmat).re;
               (*z).im=(*zmat).im;
               zmat+=1;
            }            
         }
         
         saddr=npr[ifs];
         raddr=npr[ifr];
         sbuf=(double*)(sndbuf[ifs]);
         rbuf=(double*)(rcvbuf[ifr]);

         if (np==0)
         {
            MPI_Send(sbuf,nbuf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
            MPI_Recv(rbuf,nbuf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
         }
         else
         {
            MPI_Recv(rbuf,nbuf,MPI_DOUBLE,raddr,tag,MPI_COMM_WORLD,&stat);
            MPI_Send(sbuf,nbuf,MPI_DOUBLE,saddr,tag,MPI_COMM_WORLD);
         }
      }
   }
}


static void sum_vprod(int n)
{
   int i;
   
   if (NPROC>1)
   {
      MPI_Reduce((double*)(cs1),(double*)(cs2),2*n,MPI_DOUBLE,
                 MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast((double*)(cs2),2*n,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
   else
   {
      for (i=0;i<n;i++)
      {
         cs2[i].re=cs1[i].re;
         cs2[i].im=cs1[i].im;
      }
   }
}


static void init_vectors(void)
{
   int Ns,nvh,i,j;
   complex_dble z;

   dfl_global_modes(vm);

   Ns=Nss;
   nvh=nvhs;

   for (i=0;i<Ns;i++)
   {
      set_v2zero(nvh,vs[i]+nvh);
      set_v2zero(nvh,zvs[i]+nvh);
      set_vd2zero(nvh,vds[i]+nvh);
      set_vd2zero(nvh,zvds[i]+nvh);      

      assign_v2vd(nvh,vm[i],vds[i]);
      
      for (j=0;j<i;j++)
         cs1[j]=vprod_dble(nvh,0,vds[j],vds[i]);

      sum_vprod(i);

      for (j=0;j<i;j++)
      {
         z.re=-cs2[j].re;
         z.im=-cs2[j].im;

         mulc_vadd_dble(nvh,vds[i],vds[j],z);
      }
         
      vnormalize_dble(nvh,1,vds[i]);

      Zhat_dble(vds[i],zvds[i]);
      assign_vd2v(nvh,vds[i],vs[i]);
      assign_vd2v(nvh,zvds[i],zvs[i]);
   }
}


static void set_little_matrix(void)
{
   int Ns,nvh,i,j,ie;
   double k;
   complex_dble *a;

   Ns=Nss;
   nvh=nvhs;
   a=cs1;
   
   for (i=0;i<Ns;i++)
   {
      for (j=0;j<Ns;j++)
      {
         (*a)=vprod_dble(nvh,0,vds[i],zvds[j]);
         a+=1;
      }
   }

   sum_vprod(Ns*Ns);
   ie=cmat_inv_dble(Ns,cs2,ads,&k);

   error_root((ie!=0)||(k>1.0e3),1,"set_little_matrix [Zops.c]",
              "Matrix inversion failed or was not safe\n"
              "(error code = %d, condition number = %.1e)",ie,k);
}


zop_dble_t *zophat_dble(void)
{
   int nb,Ns,nmat;
   int ib,ic,ie,is,ifc;
   double k,kmax;
   complex_dble *zee0,*zoe0,*zeo0;
   complex_dble *zee1,*zoo1,*zoe1,*zeo1;
   
   if (query_flags(ZOPHATDBLE_UP2DATE)==1)
      return &zop1;

   zopnohat_dble();
   alloc_zop1();
   nb=zop1.nb;
   Ns=zop1.Ns;
   nmat=Ns*Ns;

   is=0;
   kmax=1.0;   
   zee0=zop0.zee;
   zee1=zop1.zee;

   for (ib=0;ib<nb;ib++)
   {
      ie=cmat_inv_dble(Ns,zee0,zee1,&k);

      if ((ie!=0)&&(is==0))
         is=ie;
      
      if (k>kmax)
         kmax=k;

      zee0+=nmat;
      zee1+=nmat;
   }

   error((is!=0)||(kmax>1.0e3),1,"zophat_dble [Zops.c]",
         "Matrix inversions failed or were not safe\n"
         "(error code = %d, maximal condition number = %.1e)",is,kmax);

   if (NPROC>1)
      copy_zee();

   zoo1=zop1.zoo;
   zoe0=zop0.zoe;
   zeo0=zop0.zeo;
   zoe1=zop1.zoe;
   zeo1=zop1.zeo;
   
   for (ib=(nb/2);ib<nb;ib++)
   {
      for (ifc=0;ifc<8;ifc++)
      {
         cmat_mul_dble(Ns,zoo1,zoe0,zoe1);
         zoe0+=nmat;
         zoe1+=nmat;
         
         ic=inn[ib][ifc];

         if (ic<nb)
            zee1=zop1.zee+ic*nmat;
         else
            zee1=rcvbuf[ifc]+(ic-nb-ofsbbe[ifc])*nmat;
            
         cmat_mul_dble(Ns,zee1,zeo0,zeo1);         
         zeo0+=nmat;
         zeo1+=nmat;
      }

      zoo1+=nmat;      
   }

   set_flags(COMPUTED_ZOPHATDBLE);

   init_vectors();
   set_little_matrix(); 
   
   return &zop1;   
}


zop_t *zopnohat(void)
{
   int nb,Ns;
   complex *z2;
   complex_dble *z0,*zm;
   
   if (query_flags(ZOPNOHAT_UP2DATE)==1)
      return &zop2;

   zopnohat_dble();
   alloc_zop2();
   nb=zop2.nb;
   Ns=zop2.Ns;

   z0=zop0.zee;
   z2=zop2.zee;
   zm=zop0.zee+9*nb*Ns*Ns;

   for (;z0<zm;z0++)
   {
      (*z2).re=(float)((*z0).re);
      (*z2).im=(float)((*z0).im);
      z2+=1;
   }

   set_flags(COMPUTED_ZOPNOHAT);
   
   return &zop2;   
}


zop_t *zophat(void)
{
   int nb,Ns;
   complex *z3;
   complex_dble *z1,*zm;

   if (query_flags(ZOPHAT_UP2DATE)==1)
      return &zop3;

   zophat_dble();
   alloc_zop3();
   nb=zop3.nb;
   Ns=zop3.Ns;

   z1=zop1.zee;
   z3=zop3.zee;
   zm=zop1.zee+9*nb*Ns*Ns;

   for (;z1<zm;z1++)
   {
      (*z3).re=(float)((*z1).re);
      (*z3).im=(float)((*z1).im);
      z3+=1;
   }

   set_flags(COMPUTED_ZOPHAT);

   return &zop3;   
}


void dfl_little_modes(complex **v,complex **zv,complex_dble **vd,
                      complex_dble **zvd,complex_dble *a)
{
   int Ns,i;
   complex_dble *b,*am;

   if (query_flags(ZOPHATDBLE_UP2DATE)!=1)
      zophat_dble();

   Ns=Nss;

   for (i=0;i<Ns;i++)
   {
      v[i]=vs[i];
      zv[i]=zvs[i];
      vd[i]=vds[i];
      zvd[i]=zvds[i];
   }

   am=a+Ns*Ns;
   b=ads;
   
   for (;a<am;a++)
   {
      (*a).re=(*b).re;
      (*a).im=(*b).im;
      b+=1;
   }   
}


void free_zops(void)
{
   if (zop0.zee!=NULL)
      afree(zop0.zee);
   if (zop1.zee!=NULL)
      afree(zop1.zee);
   if (zop2.zee!=NULL)
      afree(zop2.zee);   
   if (zop3.zee!=NULL)
      afree(zop3.zee);   

   if (Nss>0)
   {
      afree(bbs[0].sdl[0]);
      afree(bbs[0].sdl);
      afree(bbs[0].ippl);

      afree(vs[0]);
      afree(vds[0]);
      afree(vs);
      afree(vds);
   }
   
   zop0.nb=0;
   zop0.nbb=0;
   zop0.Ns=0;
   zop0.zee=NULL;
   zop0.zoo=NULL;
   zop0.zoe=NULL;
   zop0.zeo=NULL;

   zop1.nb=0;
   zop1.nbb=0;
   zop1.Ns=0;
   zop1.zee=NULL;
   zop1.zoo=NULL;
   zop1.zoe=NULL;
   zop1.zeo=NULL;   
   
   zop2.nb=0;
   zop2.nbb=0;
   zop2.Ns=0;
   zop2.zee=NULL;
   zop2.zoo=NULL;
   zop2.zoe=NULL;
   zop2.zeo=NULL;

   zop3.nb=0;
   zop3.nbb=0;
   zop3.Ns=0;
   zop3.zee=NULL;
   zop3.zoo=NULL;
   zop3.zoe=NULL;
   zop3.zeo=NULL;   

   Nss=0;   
   bsv[0]=0;
   bsv[1]=0;
   bsv[2]=0;
   bsv[3]=0;
   
   set_flags(FREED_ZOPS);
}
