
/*******************************************************************************
*
* File dfl_sap_gcr.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Deflated SAP+GCR solver for the Wilson-Dirac equation (D_w+m0)*psi=eta
*
* The externally accessible functions is
*
*   double dfl_sap_gcr(int nmx,double res,int k,int l,int *status,int *ifail)
*     Obtains an approximate solution psi=psd[l] of the Wilson-Dirac equation
*     for a given source eta=psd[k], using at most nmx deflated GCR iterations,
*     with a relative residue smaller than res. The global swd field and the
*     fields on the DFL_BLOCKS and GCR_BLOCKS grids are updated automatically
*     when needed.
*      The parameter status is assumed to be an array with 3 elements. On 
*     exit status[0] reports the number of iterations that were required, or
*     a negative number if the solver failed, while status[1] and status[2]
*     report the average numbers of iterations required for the solution of 
*     the little Dirac equation in RL_pcnd() and RL_pcnd_dble(). The last
*     parameter, ifail, is set to 0 or 1 depending on whether the inversions
*     of the SW-term (if any) were safe or not. The return value is the norm
*     of the residue of the solution
*
* Notes:
*
* Evidently the SAP+GCR solver is a global program that must be called on
* all processes simultaneously. It requires a workspace of 2*nkv+1 single-
* precision spinor fields and 3 double-precision spinor fields. The workspace
* is reserved internally using the utilities in the wspace.c module (these
* must therefore also be used in the calling programs). It is permissible
* to set k=l in which case the program overwrites the source field with the
* solution (the source field is preserved if k!=l)
*
* Before running this program, the lattice, deflation and solver parameters
* must be set by calling the programs set_lat_parms(), set_gcr_parms() and
* set_dfl_parms() [flags/parms.c]. Furthermore, the deflation subspace must
* have been properly initialized by calling dfl_subspace(), dfl_modes() or
* dfl_update() (the program checks this by querying the flags data base).
* All other preparatory steps are carried out automatically when needed
*
*******************************************************************************/

#define DFL_SAP_GCR_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "start.h"
#include "block.h"
#include "linalg.h"
#include "linsolv.h"
#include "sw_term.h"
#include "dirac.h"
#include "sap_gcr.h"
#include "dfl.h"
#include "global.h"

static int nmr,ncy;
static int nit1,nit2,stat1,stat2;
static double vres,vdres,sres0,sres1,sdres0;


static void Op_dble(spinor_dble *pk,spinor_dble *pl)
{
   int k,l,status;

   k=(pk-psd[0][0])/NSPIN;
   l=(pl-psd[0][0])/NSPIN;

   Qnohat_dble(k,l);
   mulg5_dble(VOLUME,pl);

   dfl_RLpcnd_dble(k,l,&status);

   if (stat2>=0)
   {
      if (status>=0)
      {
         nit2+=1;
         stat2+=status;
      }
      else
         stat2=status;
   }
}


static void Opz(spinor *pr,spinor *paz)
{
   spinor *rsm;
   double sm;

   sm=0.0;
   rsm=pr+VOLUME;

   for (;pr<rsm;pr++)
   {
      sm+=(double)(_vector_prod_re((*pr).c1,(*pr).c1)+
                   _vector_prod_re((*pr).c2,(*pr).c2)+
                   _vector_prod_re((*pr).c3,(*pr).c3)+
                   _vector_prod_re((*pr).c4,(*pr).c4));

      _vector_sub((*paz).c1,(*pr).c1,(*paz).c1);
      _vector_sub((*paz).c2,(*pr).c2,(*paz).c2);
      _vector_sub((*paz).c3,(*pr).c3,(*paz).c3);
      _vector_sub((*paz).c4,(*pr).c4,(*paz).c4);

      paz+=1;
   }

   if (NPROC>1)
   {
      MPI_Reduce(&sm,&sres1,1,MPI_DOUBLE,MPI_SUM,0,MPI_COMM_WORLD);
      MPI_Bcast(&sres1,1,MPI_DOUBLE,0,MPI_COMM_WORLD);
   }
   else
      sres1=sm;

   sres1=sqrt(sres1);
}


static void Mj(int j,spinor *pr,spinor *pz,spinor *paz)
{
   int n,isz,isaz,status;
   double r1,r2,r3;

   isz=(pz-ps[0][0])/NSPIN;
   isaz=(paz-ps[0][0])/NSPIN;

   set_s2zero(VOLUME,pz);
   assign_s2s(VOLUME,pr,paz);

   for (n=0;n<ncy;n++)
      msap(nmr,isaz,isz);

   Opz(pr,paz);

   if (j==0)
      sres0=sres1;

   if (sres1!=0.0)
   {
      r1=1.0/sres1;
      r2=vres*sres0*r1;
      r3=vdres*sdres0*r1;

      if (r2>0.1)
         r2=0.1;
      if (r3>0.1)
         r3=0.1;
      
      set_dfl_residues(r2,r3);
   }
   
   dfl_RLpcnd(isz,isaz,&status);   

   if (stat1>=0)
   {
      if (status>=0)
      {
         nit1+=1;
         stat1+=status;
      }
      else
         stat1=status;
   }
}


double dfl_sap_gcr(int nmx,double res,int k,int l,int *status,int *ifail)
{
   int iprms[3],nb_dfl,nb_gcr,isw;
   int nkv,ird0,ird1,iwd0,lstat;
   double dprms[1],rho;
   complex_dble z;
   spinor **ws;
   spinor_dble **rsd,**wsd;
   lat_parms_t lat;
   dfl_parms_t dfl;
   gcr_parms_t gcr;

   if (NPROC>1)
   {
      iprms[0]=nmx;
      iprms[1]=k;
      iprms[2]=l;

      dprms[0]=res;

      MPI_Bcast(iprms,3,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(dprms,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

      error((iprms[0]!=nmx)||(iprms[1]!=k)||(iprms[2]!=l)||(dprms[0]!=res),1,
            "dfl_sap_gcr [dfl_sap_gcr.c]","Parameters are not global");
   }

   lat=lat_parms();
   gcr=gcr_parms();
   dfl=dfl_parms();
   
   error_root(lat.kappa==0.0,1,
              "dfl_sap_gcr [dfl_sap_gcr.c]","Lattice parameters are not set");
   error_root(gcr.nb==0,1,
              "dfl_sap_gcr [dfl_sap_gcr.c]","GCR parameters are not set");
   error_root(dfl.Ns==0,1,
              "dfl_sap_gcr [dfl_sap_gcr.c]","Deflation parameters are not set");
   error_root(query_flags(DFLMODES_UP2DATE)!=1,1,"dfl_sap_gcr [dfl_sap_gcr.c]",
              "The deflation subspace is not allocated or is not up-to-date");

   vres=dfl.vres;
   vdres=dfl.vdres;
   nmr=gcr.nmr;
   ncy=gcr.ncy;
   nkv=gcr.nkv;
   blk_list(DFL_BLOCKS,&nb_dfl,&isw);
   blk_list(GCR_BLOCKS,&nb_gcr,&isw);      

   ws=reserve_ws(2*nkv+1);
   rsd=reserve_wsd(2);
   ird0=(rsd[0]-psd[0][0])/NSPIN;
   ird1=(rsd[1]-psd[0][0])/NSPIN;

   error_root((k<0)||(k>=no_sd)||(k==ird0)||(k==ird1)||
              (l<0)||(l>=no_sd)||(l==ird0)||(l==ird1),1,
              "dfl_sap_gcr [dfl_sap_gcr.c]",
              "Improper workspace or arguments k,l");
   
   if ((query_flags(SWD_UP2DATE)!=1)||
       (query_flags(SWD_E_INVERTED)!=0)||(query_flags(SWD_O_INVERTED)!=0))
   {
      sw_term();
   }

   if (query_flags(U_MATCH_UD)!=1)
   {
      assign_ud2u();
   }

   if ((query_flags(SW_UP2DATE)!=1)||(query_flags(SW_E_INVERTED)!=0)||
       (query_flags(SW_O_INVERTED)!=0))
   {
      assign_swd2sw();
   }
   
   if (query_blk_flags(GCR_BLOCKS,UBLK_MATCH_UD,nb_gcr)!=1)
   {
      assign_ud2ubgr(GCR_BLOCKS);
   }

   if ((query_blk_flags(GCR_BLOCKS,SW_UP2DATE,nb_gcr)!=1)||
       (query_blk_flags(GCR_BLOCKS,SW_E_INVERTED,nb_gcr)!=0)||
       (query_blk_flags(GCR_BLOCKS,SW_O_INVERTED,nb_gcr)!=1))
   {
      (*ifail)=assign_swd2swbgr(GCR_BLOCKS,ODD_PTS);
   }
   else
      (*ifail)=0;

   rho=norm_square_dble(VOLUME,1,psd[k][0]);
   rho=sqrt(rho);
   
   if ((*ifail)==0)
   {
      assign_sd2sd(VOLUME,psd[k][0],rsd[0]);
      dfl_Lpcnd_dble(ird0,ird1,&lstat);
      sdres0=norm_square_dble(VOLUME,1,rsd[0]);
      sdres0=sqrt(sdres0);

      if ((lstat>=0)&&((res*rho)<sdres0))
      {
         nit1=0;
         nit2=0;
         stat1=0;
         stat2=0;

         wsd=reserve_wsd(1);         
         iwd0=(wsd[0]-psd[0][0])/NSPIN;   
         error_root((k==iwd0)||(l==iwd0),1,"dfl_sap_gcr [dfl_sap_gcr.c]",
                    "Improper workspace or arguments k,l");

         res=(res*rho)/sdres0;
         rho=fgcr(VOLUME,1,Op_dble,Mj,ws,wsd,nkv,nmx,res,
                  rsd[0],psd[l][0],status);

         release_wsd();
         z.re=1.0;
         z.im=0.0;
         mulc_spinor_add_dble(VOLUME,psd[l][0],rsd[1],z);

         if (stat1>0)
            status[1]=stat1/nit1;
         else
            status[1]=stat1;

         if (stat2>0)
            status[2]=stat2/nit2;
         else
            status[2]=stat2;
      }
      else
      {
         status[0]=0;
         status[1]=0;
         status[2]=lstat;

         if (lstat>=0)
         {
            assign_sd2sd(VOLUME,rsd[1],psd[l][0]);
            rho=sdres0;
         }
         else
            set_sd2zero(VOLUME,psd[l][0]);
      }
   }
   else
   {
      status[0]=0;
      status[1]=0;
      status[2]=0;
      set_sd2zero(VOLUME,psd[l][0]);      
   }
   
   set_dfl_residues(vres,vdres);
   release_wsd();
   release_ws();

   return rho;
}
