
/*******************************************************************************
*
* File vcom.c
*
* Copyright (C) 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Communication functions for the global single-precision vector fields
*
*   void cpv_int_bnd(complex *v)
*     Copies the components of the field v[] on the interior boundary of
*     the local block lattice to the corresponding field components at
*     the exterior boundaries of the block lattices on the neighbouring
*     processes
*
*   void cpv_ext_bnd(complex *v)
*     *Adds* the components of the field v[] on the exterior boundary of
*     the local block lattice to the corresponding field components at
*     the interior boundaries of the block lattices on the neighbouring
*     processes
*
*   void free_vbufs(void)
*     Frees the communication buffers used by the programs in this module
* 
* Notes:
*
* The vector fields allocated by the programs in the module vflds.c are
* defined on the local DFL_BLOCKS lattices and the *even* blocks on the
* exterior boundaries of these. Accordingly, the communication programs
* only copy the field components associated to the even blocks on the
* exterior boundaries and the odd blocks on the interior boundaries
*
* The copy programs allocate the required communication buffers when
* needed. They can be freed using free_vbufs(), but if the programs are
* called frequently, it is better to leave the buffers allocated (in which
* case they will be reused)
*
* All these programs act globally. In particular, the copy programs assume
* that v is the starting address of one of the globally defined fields.
* This will normally be guaranteed by the calling program so that no check
* is made here
*
*******************************************************************************/

#define VCOM_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "dfl.h"
#include "global.h"

static int np,nmu[8];
static int nb,Ns;
static int *ipp,nbbe[8],nbbo[8],ofsbbe[8],ofsbbo[8];
static int nfc_int,ifc_int[8];
static int nfc_ext,ifc_ext[8];
static int itags=0,tags[16];
static complex *snd_buf_int[8],*rcv_buf_int[8];
static complex *snd_buf_ext[8],*rcv_buf_ext[8],*wb=NULL;
static MPI_Request snd_req_int[8],rcv_req_int[8];
static MPI_Request snd_req_ext[8],rcv_req_ext[8];


static void get_tags(void)
{
   int i;
   
   if (itags==0)
   {
      for (i=0;i<16;i++)
         tags[i]=mpi_permanent_tag();

      itags=1;
   }
}
   

static void alloc_vbufs(void)
{
   int isw,nbb,n;
   int ifc,tag,saddr,raddr;
   complex *wbb;
   block_t *b;
   dfl_parms_t dfl;

   np=(cpr[0]+cpr[1]+cpr[2]+cpr[3])&0x1;

   for (ifc=0;ifc<8;ifc++)
      nmu[ifc]=cpr[ifc/2]&0x1;
   
   b=blk_list(DFL_BLOCKS,&nb,&isw);
   error(b==NULL,1,"alloc_vbufs [vcom.c]","DFL_BLOCKS grid is not allocated");

   dfl=dfl_parms();
   Ns=dfl.Ns;
   ipp=vipp(nbbe,nbbo,ofsbbe,ofsbbo);
   error(ipp==NULL,1,"alloc_vbufs [vcom.c]",
         "Geometry arrays associated to the DFL_BLOCKS grid are not set");
   
   nbb=0;

   for (ifc=0;ifc<8;ifc++)
      nbb+=(nbbe[ifc]+nbbo[ifc]);
   
   wb=amalloc(nbb*Ns*sizeof(*wb),ALIGN);
   error(wb==NULL,1,"alloc_vbufs [vcom.c]",
         "Unable to allocate communication buffers");

   get_tags();
   n=0;
   wbb=wb;

   for (ifc=0;ifc<8;ifc++)
   {
      if (nbbo[ifc]>0)
      {
         ifc_int[n]=ifc;
         n+=1;

         snd_buf_int[ifc]=wbb;
         wbb+=nbbo[ifc]*Ns;
         rcv_buf_int[ifc]=wbb;
         wbb+=nbbe[ifc^0x1]*Ns;

         tag=tags[ifc];
         saddr=npr[ifc];
         raddr=npr[ifc^0x1];

         MPI_Send_init((float*)(snd_buf_int[ifc]),2*nbbo[ifc]*Ns,
                       MPI_FLOAT,saddr,tag,MPI_COMM_WORLD,&snd_req_int[ifc]);
         MPI_Recv_init((float*)(rcv_buf_int[ifc]),2*nbbe[ifc^0x1]*Ns,
                       MPI_FLOAT,raddr,tag,MPI_COMM_WORLD,&rcv_req_int[ifc]);
      }
   }

   nfc_int=n;
   
   n=0;
   wbb=wb;

   for (ifc=0;ifc<8;ifc++)
   {
      if (nbbe[ifc]>0)
      {
         ifc_ext[n]=ifc;
         n+=1;

         snd_buf_ext[ifc]=wbb;
         wbb+=nbbe[ifc]*Ns;
         rcv_buf_ext[ifc]=wbb;
         wbb+=nbbo[ifc^0x1]*Ns;

         tag=tags[ifc+8];         
         saddr=npr[ifc];
         raddr=npr[ifc^0x1];

         MPI_Send_init((float*)(snd_buf_ext[ifc]),2*nbbe[ifc]*Ns,
                       MPI_FLOAT,saddr,tag,MPI_COMM_WORLD,&snd_req_ext[ifc]);
         MPI_Recv_init((float*)(rcv_buf_ext[ifc]),2*nbbo[ifc^0x1]*Ns,
                       MPI_FLOAT,raddr,tag,MPI_COMM_WORLD,&rcv_req_ext[ifc]);
      }
   }

   nfc_ext=n;   
}


static void get_int(int n,int *imb,complex *v,complex *w)
{
   int *imm;
   complex *vv,*vm;

   imm=imb+n;
   
   for (;imb<imm;imb+=1)
   {
      vv=v+(*imb)*Ns;
      vm=vv+Ns;
      
      for (;vv<vm;vv++)
      {
         (*w).re=(*vv).re;
         (*w).im=(*vv).im;
         w+=1;
      }
   }
}


static void send_bufs_int(int ifc,int eo)
{
   int io;

   if (np==eo)
   {
      io=(ifc^nmu[ifc]);
      
      MPI_Start(&snd_req_int[io]);
   }
   else
   {
      io=(ifc^nmu[ifc])^0x1;
      
      MPI_Start(&rcv_req_int[io]);
   }
}


static void wait_bufs_int(int ifc,int eo)
{
   int io;
   MPI_Status stat_snd,stat_rcv;

   if (np==eo)
   {
      io=(ifc^nmu[ifc]);
      
      MPI_Wait(&snd_req_int[io],&stat_snd);
   }
   else
   {
      io=(ifc^nmu[ifc])^0x1;
      
      MPI_Wait(&rcv_req_int[io],&stat_rcv);      
   }   
}


void cpv_int_bnd(complex *v)
{
   int ifc,io;
   int n,m,eo;

   if (BNDRY==0)
      return;
   else if (wb==NULL)
      alloc_vbufs();

   m=0;
   eo=0;

   for (n=0;n<nfc_int;n++)
   {
      if (n>0)
         send_bufs_int(ifc_int[m],eo);

      ifc=ifc_int[n];
      io=ifc^nmu[ifc];
      get_int(nbbo[io],ipp+ofsbbo[io],v,snd_buf_int[io]);

      if (n>0)
      {
         wait_bufs_int(ifc_int[m],eo);
         m+=eo;
         eo^=0x1;
      }
   }

   for (n=0;n<2;n++)
   {
      send_bufs_int(ifc_int[m],eo);
      wait_bufs_int(ifc_int[m],eo);
      m+=eo;
      eo^=0x1;
   }
   
   for (n=0;n<nfc_int;n++)
   {
      if (m<nfc_int)
         send_bufs_int(ifc_int[m],eo);

      ifc=ifc_int[n];
      io=(ifc^nmu[ifc])^0x1;
      assign_v2v(nbbe[io^0x1]*Ns,rcv_buf_int[io],v+(nb+ofsbbe[io^0x1])*Ns);

      if (m<nfc_int)
      {
         wait_bufs_int(ifc_int[m],eo);
         m+=eo;
         eo^=0x1;
      }
   }
}


static void put_ext(int n,int *imb,complex *w,complex *v)
{
   int *imm;
   complex *vv,*vm;

   imm=imb+n;
   
   for (;imb<imm;imb+=1)
   {
      vv=v+(*imb)*Ns;
      vm=vv+Ns;
      
      for (;vv<vm;vv++)
      {
         (*vv).re+=(*w).re;
         (*vv).im+=(*w).im;
         w+=1;
      }
   }
}


static void send_bufs_ext(int ifc,int eo)
{
   int io;

   if (np==eo)
   {
      io=(ifc^nmu[ifc]);
      
      MPI_Start(&snd_req_ext[io]);
   }
   else
   {
      io=(ifc^nmu[ifc])^0x1;
      
      MPI_Start(&rcv_req_ext[io]);
   }
}


static void wait_bufs_ext(int ifc,int eo)
{
   int io;
   MPI_Status stat_snd,stat_rcv;

   if (np==eo)
   {
      io=(ifc^nmu[ifc]);
      
      MPI_Wait(&snd_req_ext[io],&stat_snd);
   }
   else
   {
      io=(ifc^nmu[ifc])^0x1;
      
      MPI_Wait(&rcv_req_ext[io],&stat_rcv);      
   }   
}


void cpv_ext_bnd(complex *v)
{
   int ifc,io;
   int n,m,eo;

   if (BNDRY==0)
      return;
   else if (wb==NULL)
      alloc_vbufs();

   m=0;
   eo=0;
   
   for (n=0;n<nfc_ext;n++)
   {
      if (n>0)
         send_bufs_ext(ifc_ext[m],eo);

      ifc=ifc_ext[n];
      io=ifc^nmu[ifc];
      assign_v2v(nbbe[io]*Ns,v+(nb+ofsbbe[io])*Ns,snd_buf_ext[io]);      

      if (n>0)
      {
         wait_bufs_ext(ifc_ext[m],eo);
         m+=eo;
         eo^=0x1;
      }
   }

   for (n=0;n<2;n++)
   {
      send_bufs_ext(ifc_ext[m],eo);
      wait_bufs_ext(ifc_ext[m],eo);
      m+=eo;
      eo^=0x1;
   }
   
   for (n=0;n<nfc_ext;n++)
   {
      if (m<nfc_ext)
         send_bufs_ext(ifc_ext[m],eo);

      ifc=ifc_ext[n];
      io=(ifc^nmu[ifc])^0x1;
      put_ext(nbbo[io^0x1],ipp+ofsbbo[io^0x1],rcv_buf_ext[io],v);

      if (m<nfc_ext)
      {
         wait_bufs_ext(ifc_ext[m],eo);
         m+=eo;
         eo^=0x1;
      }
   }
}


void free_vbufs(void)
{
   int n,ifc;

   if (wb==NULL)
      return;

   for (n=0;n<nfc_int;n++)
   {
      ifc=ifc_int[n];
      MPI_Request_free(&snd_req_int[ifc]);
      MPI_Request_free(&rcv_req_int[ifc]);
   }

   for (n=0;n<nfc_ext;n++)
   {
      ifc=ifc_ext[n];
      MPI_Request_free(&snd_req_ext[ifc]);
      MPI_Request_free(&rcv_req_ext[ifc]);
   }   

   afree(wb);
   wb=NULL;
}
