
/*******************************************************************************
*
* File bshift.c
*
* Copyright (C) 2006 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Block shift generator 
*
* The externally accessible functions are
*
*   int start_bshift(*bs)
*     Initializes the block shift generator for the specified block size.
*     The block sizes bs[4] must be even and not smaller than 4. On exit
*     the progam returns the cycle length of the generator
*
*   void next_bshift(*svec)
*     Computes the next shift vector svec[4]. This program assumes that
*     the generator was previously initialized using start_bshift()
*
* Notes:
*
* The generator is the one described in appendix C of
*
*  M. Luescher: Comp. Phys. Comm. 165 (2005) 199
*
* It returns a uniformly distributed random vector followed by n-1 fixed
* shift vectors along the principal diagonal of the block. Then again a
* random vector, and so on. The generator thus has a cycle of length n
* where
*
*   n = min{10,bs[0],bs[1],bs[2],bs[3]}/2
*
* depends on the block size. The generated shift vectors svec[4] are
* guaranteed to be the same on all processes, and their components are
* in the range 0<=svec[mu]<bs[mu].
*
* When start_bshift() is called a second time, the block shift generator
* is restarted, using the specified block size.
*
* These programs perform global operations and should therefore be called
* simultaneously on all processes. 
*
*******************************************************************************/

#define BSHIFT_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "random.h"
#include "global.h"

static int b[4],init=0;
static int nvec,ivec,vec[4][4];


int start_bshift(int *bs)
{
   int iprms[4];
   int mu,i,v,r;

   if (NPROC>1)
   {
      iprms[0]=bs[0];
      iprms[1]=bs[1];
      iprms[2]=bs[2];
      iprms[3]=bs[3];
      
      MPI_Bcast(iprms,4,MPI_INT,0,MPI_COMM_WORLD);

      error((iprms[0]!=bs[0])||(iprms[1]!=bs[1])||
            (iprms[2]!=bs[2])||(iprms[3]!=bs[3]),1,
            "start_bshift [bshift.c]","The block sizes are not global");
   }

   error_root(((bs[0]&0x1)!=0)||(bs[0]<4)||((bs[1]&0x1)!=0)||(bs[1]<4)||
              ((bs[2]&0x1)!=0)||(bs[2]<4)||((bs[3]&0x1)!=0)||(bs[3]<4),1,
              "start_bshift [bshift.c]",
              "The block sizes must be even and not smaller than 4");
                 
   nvec=10;
      
   for (mu=0;mu<4;mu++)
   {
      b[mu]=bs[mu];

      if (b[mu]<nvec)
         nvec=b[mu];
   }

   nvec/=2;

   for (mu=0;mu<4;mu++)
   {
      v=b[mu]/nvec;
      r=b[mu]-nvec*v;

      if ((2*r)<nvec)
      {
         for (i=0;i<(nvec-1);i++)
         {
            if (((i&0x1)==1)&&(r>0))
            {
               vec[i][mu]=v+1;
               r-=1;
            }
            else
               vec[i][mu]=v;
         }            
      }
      else         
      {
         v+=1;
         r-=nvec;

         for (i=0;i<(nvec-1);i++)
         {
            if (((i&0x1)==1)&&(r<0))
            {
               vec[i][mu]=v-1;
               r+=1;
            }
            else
               vec[i][mu]=v;
         }             
      }
   }

   ivec=nvec-1;
   init=1;

   return nvec;
}


static void random_vec(int *svec)
{
   int mu;
   double r[4];

   ranlxd(r,4);
   
   for (mu=0;mu<4;mu++)
   {
      svec[mu]=(int)((double)(b[mu])*r[mu]);

      if (svec[mu]>=b[mu])
         svec[mu]=b[mu]-1;
   }
}


void next_bshift(int *svec)
{
   int mu;

   error(init==0,1,"next_bshift [bshift.c]",
         "The block shift generator is not initialized");
      
   if (ivec==(nvec-1))
   {
      random_vec(svec);
      MPI_Bcast(svec,4,MPI_INT,0,MPI_COMM_WORLD);
      ivec=0;
   }
   else
   {
      for (mu=0;mu<4;mu++)
         svec[mu]=vec[ivec][mu];

      ivec+=1;
   }
}

