
/*******************************************************************************
*
* File force1.c
*
* Copyright (C) 2005, 2007 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Calculation of the force F_1 that derives from the block pseudo-fermion
* action and the determinants of the SW-term
*
* The externally accessible functions are
*
*   void hmc_frf(int n,int *status,int *ifail)
*     Computes the force on the n'th block in the HMC_BLOCKS grid. On exit
*     status reports the  number of CG iterations that were required in this
*     calculation or a negative value if the program failed. The SW-term
*     is recomputed and inverted when needed and ifail is set to 0 or 1
*     depending on whether the inversions were safe or not
*
*   double hmc_acf(int n,int iphi,int idet,int *status,int *ifail)
*     Returns the pseudo-fermion action on the n'th block of the HMC_BLOCKS
*     grid (if iphi=1) plus the contribution of the determinants of the
*     SW-term to the block action (if idet=1). On exit status reports the 
*     number of BiCG iterations that were required in this calculation or
*     a negative value if the program failed. The SW-term is recomputed 
*     and inverted when needed and ifail is set to 0 or 1 depending on
*     whether the inversions were safe or not
*
* Notes:
*
* The programs in this file do not require any communications. Whether the
* SW-term on the blocks is in the proper condition is checked by querying
* the flags data base, and the proper action is then taken. Any errors
* are recorded by the error_loc() function
*
* The lattice parameters, maximal iteration numbers, ... are retrieved
* from the parameter data base using the utilities provided by the module
* flags/parms.c. Only the forces on the active links are calculated, and 
* the force field on inactive links is left untouched. The "chronological
* inversion method" is silently applied in the program hmc_frf() if the 
* parameter hmc.cno_blk returned by hmc_parms() is not equal to zero
*
* Note that the force field is a shared field. Moreover the same field is
* used for the various forces (gauge, block, block interaction)
*
*******************************************************************************/

#define FORCE1_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "start.h"
#include "misc.h"
#include "linalg.h"
#include "linsolv.h"
#include "eva.h"
#include "block.h"
#include "sw_term.h"
#include "dirac.h"
#include "hmc.h"
#include "global.h"

#define DEBUG 0
#define MAX_LEVELS 8
#define BLK_LENGTH 8

static int nmat=0,nold;
static int cnt[MAX_LEVELS];
static double smx[MAX_LEVELS],*rdold;
static complex_dble *amat,*bmat,*cvec,*dvec;
static spinor_dble **sdold;
static block_t *bsv;


static void alloc_mat(int n)
{
   if (n<=nmat)
      return;

   if (nmat>0)
   {
      afree(rdold);
      afree(amat);
      afree(sdold);
   }

   rdold=amalloc(n*sizeof(*rdold),3);
   amat=amalloc(2*n*(n+1)*sizeof(*amat),ALIGN);
   sdold=amalloc(n*sizeof(*sdold),3);
   
   if (error_loc((rdold==NULL)||(amat==NULL)||(sdold==NULL),1,
                 "alloc_mat [force1.c]","Unable to allocate auxiliary arrays"))
   {
      if (rdold!=NULL)
         afree(rdold);
      if (amat!=NULL)
         afree(amat);
      if (sdold!=NULL)
         afree(sdold);

      rdold=NULL;
      amat=NULL;
      bmat=NULL;
      cvec=NULL;
      dvec=NULL;
      sdold=NULL;
      nmat=0;
   }
   else
   {
      bmat=amat+n*n;
      cvec=bmat+n*n;
      dvec=cvec+n;
      nmat=n;
   }
}


static void least_squares(void)
{
   int ie,i;
   double k;

   ie=cmat_inv_dble(nold,amat,bmat,&k);

   if ((ie==0)&&(k<1.0e8))
      cmat_vec_dble(nold,bmat,cvec,dvec);
   else
   {
      dvec[0].re=rdold[0];
      dvec[0].im=0.0;
      
      for (i=1;i<nold;i++)
      {
         dvec[i].re=0.0;
         dvec[i].im=0.0;
      }
   }
}


static void propagate_old(void)
{
   int i,j;
   double rd;
   complex_dble z;
   
   for (i=0;i<nold;i++)
   {
      assign_sd2sd((*bsv).volh,sdold[i],(*bsv).sd[0]);
      cvec[i]=spinor_prod_dble((*bsv).volh,0,(*bsv).sd[0],(*bsv).sd[1]);
      rd=1.0/rdold[i];
      cvec[i].re*=rd;
      cvec[i].im*=rd;      

      Qhat_blk_dble(bsv,0,i+2);
      rd=norm_square_dble((*bsv).volh,0,(*bsv).sd[i+2]);      
      amat[i*nold+i].re=rd/(rdold[i]*rdold[i]);
      amat[i*nold+i].im=0.0;

      for (j=0;j<i;j++)
      {
         z=spinor_prod_dble((*bsv).volh,0,(*bsv).sd[i+2],(*bsv).sd[j+2]);
         rd=1.0/(rdold[i]*rdold[j]);
         z.re*=rd;
         z.im*=rd;
         
         amat[i*nold+j].re=z.re;
         amat[i*nold+j].im=z.im;
         amat[j*nold+i].re=z.re;
         amat[j*nold+i].im=-z.im;         
      }
   }

   least_squares();      
   set_sd2zero((*bsv).volh,(*bsv).sd[0]);

   for (i=(nold-1);i>=0;i--)
   {
      rd=1.0/rdold[i];
      z.re=dvec[i].re*rd;
      z.im=dvec[i].im*rd;
      mulc_spinor_add_dble((*bsv).volh,(*bsv).sd[0],sdold[i],z);
   }
}


static void Op1(spinor *pk,spinor *pl)
{
   int k,l;

   k=(pk-(*bsv).s[0])/((*bsv).vol+1);
   l=(pl-(*bsv).s[0])/((*bsv).vol+1);
   
   Qhat_blk(bsv,k,l);
}


static void Op1_dble(spinor_dble *pk,spinor_dble *pl)
{
   int k,l;

   k=(pk-(*bsv).sd[0])/((*bsv).vol+1);
   l=(pl-(*bsv).sd[0])/((*bsv).vol+1);
   
   Qhat_blk_dble(bsv,k,l);
}


static void third_vec(int volh,spinor_dble *sl,spinor_dble *phi,spinor_dble *st)
{
   spinor_dble *sm;

   sm=sl+volh;

   for (;sl<sm;sl++)
   {
      _vector_sub((*st).c1,(*phi).c1,(*sl).c1);
      _vector_sub((*st).c2,(*phi).c2,(*sl).c2);
      _vector_sub((*st).c3,(*phi).c3,(*sl).c3);      
      _vector_sub((*st).c4,(*phi).c4,(*sl).c4);      

      phi+=1;
      st+=1;
   }

   sm=sl+volh;

   for (;sl<sm;sl++)
   {
      _vector_mul((*st).c1,-1.0,(*sl).c1);
      _vector_mul((*st).c2,-1.0,(*sl).c2);
      _vector_mul((*st).c3,-1.0,(*sl).c3);
      _vector_mul((*st).c4,-1.0,(*sl).c4);      

      st+=1;
   }
}


void hmc_frf(int n,int *status,int *ifail)
{
#if (DEBUG>=1)
   int nb,isw,i;
#else
   int nb,isw;
#endif
   spinor_dble *rb,*sb,*tb,*vb;
   double res0,res1;
   complex_dble z;
   hmc_t *hflds;
   hmc_parms_t hmc;

   (*status)=0;
   (*ifail)=0;   
   bsv=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   if (error_loc((hflds==NULL)||(n<0)||(n>=nb),1,"hmc_frf [force1.c]",
                 "Improper block grid, block fields or block number"))
      return;
   
   if ((query_blk_flags(HMC_BLOCKS,SWD_UP2DATE,n)!=1)||
       (query_blk_flags(HMC_BLOCKS,SWD_E_INVERTED,n)!=0))
      blk_sw_term(n);   

   if (query_blk_flags(HMC_BLOCKS,SWD_O_INVERTED,n)!=1)
   {
      if (invert_blk_swd(HMC_BLOCKS,n,ODD_PTS))
      {
         (*ifail)=1;
         return;
      }
   }

   assign_udb2ub(HMC_BLOCKS,n);
   assign_swdb2swb(HMC_BLOCKS,n,NO_PTS);
   
   hflds+=n;   
   bsv+=n;
   rb=(*bsv).sd[0];
   sb=(*bsv).sd[1];
   tb=(*bsv).sd[2];
   vb=(*bsv).sd[3];
   hmc=hmc_parms();

   apply_sw_dble((*bsv).volh,(*bsv).swd,(*hflds).phi,sb);
   
   if (hmc.cno_blk>0)
   {
      alloc_mat(hmc.cno_blk);

      if (nmat>0)
         nold=old_chrono_blk(n,rdold,sdold);
      else
         nold=0;
   }
   else
   {
      nmat=0;
      nold=0;
   }

#if (DEBUG>=1)
   if (n==0)
      message("In hmc_frf(): nmat= %d, nold = %d\n",
              nmat,nold);
#endif
   
   if (nold==0)
   {
      res1=1.0;
      cgne((*bsv).volh,0,Op1,Op1_dble,(*bsv).s,(*bsv).sd+2,
           hmc.nmx_frf,hmc.res_frf,sb,rb,status);
   }
   else
   {
      propagate_old();

      Qhat_blk_dble(bsv,0,2);
      Qhat_blk_dble(bsv,2,3);
      z.re=-1.0;
      z.im=0.0;
      mulc_spinor_add_dble((*bsv).volh,vb,sb,z);
      
      res0=norm_square_dble((*bsv).volh,0,sb);      
      res1=norm_square_dble((*bsv).volh,0,vb);
      res1=sqrt(res1/res0);

#if (DEBUG>=1)
      if (n==0)
      {
         message("Least squares: oldvec norm = %.2e",rdold[0]);

         for (i=1;i<nold;i++)
            message(", %.2e",rdold[i]);

         message("\nres1 = %.1e\n",res1);
      }
#endif
      
      if (res1<1.0)
      {
         if (res1>hmc.res_frf)
         {
            cgne((*bsv).volh,0,Op1,Op1_dble,(*bsv).s,(*bsv).sd+4,
                 hmc.nmx_frf,hmc.res_frf/res1,vb,tb,status);
            z.re=-1.0;
            z.im=0.0;
            mulc_spinor_add_dble((*bsv).volh,rb,tb,z);
         }
         else
            (*status)=0;
      }
      else
         cgne((*bsv).volh,0,Op1,Op1_dble,(*bsv).s,(*bsv).sd+2,
              hmc.nmx_frf,hmc.res_frf,sb,rb,status);
   } 
   
   if ((hmc.cno_blk>0)&&((*status)>0))
      add_chrono_blk(n,rb);
      
#if (DEBUG>=1)
   if (n==0)
   {
      Qhat_blk_dble(bsv,0,2);
      Qhat_blk_dble(bsv,2,3);
      z.re=-1.0;
      z.im=0.0;
      mulc_spinor_add_dble((*bsv).volh,vb,sb,z);
      res1=norm_square_dble((*bsv).volh,0,vb)/
         norm_square_dble((*bsv).volh,0,sb);
      res1=sqrt(res1);

      message("Status = %d, residue = %.1e (should be <= %.1e)\n",
              (*status),res1,hmc.res_frf);
   }
#endif        

   Qhat_blk_dble(bsv,0,1);
   Qoe_blk_dble(bsv,0,0);
   Qoe_blk_dble(bsv,1,1);

   apply_sw_dble((*bsv).volh,(*bsv).swd+(*bsv).vol,
                 rb+(*bsv).volh,rb+(*bsv).volh);   
   apply_sw_dble((*bsv).volh,(*bsv).swd+(*bsv).vol,
                 sb+(*bsv).volh,sb+(*bsv).volh);   
   
   third_vec((*bsv).volh,sb,(*hflds).phi,tb);

   if (sw_frc(n,0,2,1))
   {
      (*ifail)=1;
      return;
   }

   hop_frc(n,0,1);
}


static void Op2_dble(spinor_dble *pk,spinor_dble *pl)
{
   int k,l;

   k=(pk-(*bsv).sd[0])/((*bsv).vol+1);
   l=(pl-(*bsv).sd[0])/((*bsv).vol+1);
   
   Qhat_blk_dble(bsv,k,l);
   mulg5_dble((*bsv).volh,pl);
}


static double act_det(int vol,pauli_dble *m)
{
   int n;
   double c,p,act;
   pauli_dble *mm,*mn;
   lat_parms_t lat;

   lat=lat_parms();
   c=pow(4.0+lat.m0,-6.0);

   for (n=0;n<MAX_LEVELS;n++)
   {
      cnt[n]=0;
      smx[n]=0.0;
   }
   
   mm=m+2*vol;

   for (;m<mm;)
   {
      mn=m+BLK_LENGTH;
      if (mn>mm)
         mn=mm;
      p=1.0;

      for (;m<mn;m++)
         p*=(c*det_pauli_dble(m));

      cnt[0]+=1;
      smx[0]-=log(fabs(p));

      for (n=1;(cnt[n-1]>=BLK_LENGTH)&&(n<MAX_LEVELS);n++)
      {
         cnt[n]+=1;
         smx[n]+=smx[n-1];

         cnt[n-1]=0;
         smx[n-1]=0.0;
      }
   }

   act=0.0;

   for (n=0;n<MAX_LEVELS;n++)
      act+=smx[n];   
   
   return act+act;
}


double hmc_acf(int n,int iphi,int idet,int *status,int *ifail)
{
   int nb,isw;
   double act;
   spinor_dble *rb,*sb;
   hmc_t *hflds;
   hmc_parms_t hmc;

   (*status)=0;
   (*ifail)=0;

   if ((iphi!=1)&&(idet!=1))
      return 0.0;
   
   bsv=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   if (error_loc((hflds==NULL)||(n<0)||(n>=nb),1,"hmc_acf [force1.c]",
                 "Improper block grid, block fields or block number"))
      return 0.0;

   act=0.0;
   bsv+=n;
   hflds+=n;   

   if (idet==1)
   {
      if ((query_blk_flags(HMC_BLOCKS,SWD_UP2DATE,n)!=1)||
          (query_blk_flags(HMC_BLOCKS,SWD_E_INVERTED,n)!=0)||
          (query_blk_flags(HMC_BLOCKS,SWD_O_INVERTED,n)!=0))
         blk_sw_term(n);   

      act+=act_det((*bsv).vol,(*bsv).swd);
   }

   if (iphi==1)
   {
      if ((query_blk_flags(HMC_BLOCKS,SWD_UP2DATE,n)!=1)||
          (query_blk_flags(HMC_BLOCKS,SWD_E_INVERTED,n)!=0))
         blk_sw_term(n);   

      if (query_blk_flags(HMC_BLOCKS,SWD_O_INVERTED,n)!=1)
      {
         if (invert_blk_swd(HMC_BLOCKS,n,ODD_PTS))
         {
            (*ifail)=1;
            return 0.0;
         }
      }

      rb=(*bsv).sd[0];
      sb=(*bsv).sd[1];
   
      apply_sw_dble((*bsv).volh,(*bsv).swd,(*hflds).phi,sb);   
      mulg5_dble((*bsv).volh,sb);

      hmc=hmc_parms();   
      bicg((*bsv).volh,0,Op2_dble,(*bsv).sd+2,hmc.nmx_acf,hmc.res_acf,
           sb,rb,status);

      if ((*status)>=0)
         act+=norm_square_dble((*bsv).volh,0,rb);
   }
   
   return act; 
}

