
/*******************************************************************************
*
* File genfrc.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generic force calculation programs
*
* The externally accessible functions are
*
*   int sw_frc(int n,int k,int l,int idet)
*     Computes the force F on the n'th block in the HMC_BLOCKS grid which
*     satisfies, for any variation omega of the gauge field, 
*
*        (omega,F)=2*Re(psi_k,delta_omega(Q_ee+Q_oo)*psi_l)
*
*     where psi_k and psi_l are the k'th and the l'th double-precision block
*     spinor fields. The contribution from the "small" determinant
*
*        (omega,F)=-2*tr{delta_omega(Q_ee+Q_oo)*(Q_ee+Q_oo)^(-1)}
*
*     is added if idet=1, and the calculated force is then assigned to the
*     force field on the block. The return value is 1 if idet=1 and if the
*     inversion of some of the Pauli matrices was unsafe. Otherwise the 
*     return value is 0
*
*   void hop_frc(int n,int k,int l)
*     Computes the force F on the n'th block of the HMC_BLOCKS grid which
*     satisfies, for any variation omega of the gauge field, 
*
*        (omega,F)=2*Re(psi_k,delta_omega(Q_oe+Q_eo)*psi_l)
*
*     where psi_k and psi_l are the k'th and the l'th double-precision block
*     spinor fields. The calculated force is then *added* to the force field
*     on the block
*
* Notes:
*
* The programs in this file can be called locally. If an error is detected,
* no operations are performed and an error message is recorded by the
* error_loc function (unsafe inversions of Pauli matrices are not treated
* as an error since these can be detected by the calling program)
*
* The lattice parameters are retrieved from the parameter data base using
* the utilities provided by the module flags/parms.c. Only the forces on
* the active links are calculated, and the force field on inactive links
* is left untouched
*
* The program sw_frc() recalculates the block SW-term if idet=1 and if the
* block swd array is not up-to-date. On exit the array contains the inverted
* SW-term if idet=1 and is unchanged otherwise
*
* Note that the force field is a shared field. Moreover the same field is
* used for the various forces (gauge, block, block interaction)
*
*******************************************************************************/

#define GENFRC_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "misc.h"
#include "start.h"
#include "block.h"
#include "sw_term.h"
#include "hmc.h"
#include "global.h"

static int size=0;
static u3_alg_dble *Xfld;

#if (defined SSE2)
#include "sse2.h"
static su3_dble w1,w2,w3 __attribute__ ((aligned (16)));
static const su3_alg_dble X0 __attribute__ ((aligned (16))) ={0.0};
static su3_alg_dble Yw __attribute__ ((aligned (16)));
static u3_alg_dble Zw[6] __attribute__ ((aligned (16)));
#else
static su3_dble w1,w2,w3;
static const su3_alg_dble X0={0.0};
static su3_alg_dble Yw;
static u3_alg_dble Zw[6];
#endif


static void set_frc2zero(hmc_t *hflds)
{
   int *lk,*lkm;
   su3_alg_dble *fb;

   fb=(*hflds).frc;   
   lk=(*hflds).lks;
   lkm=lk+(*hflds).nlks;
   
   for (;lk<lkm;lk++)
      *(fb+(*lk))=X0;
}


static void det2X(int vol,double c,pauli_dble *m,u3_alg_dble *X)
{
   pauli_dble *mm;
   u3_alg_dble *Xm,*Y;

   mm=m+2*vol;

   for (;m<mm;m+=2)
   {
      det2u3alg(m,Zw);

      Y=Zw;
      Xm=X+6;
      
      for (;X<Xm;X++)
      {         
         (*X).c1+=c*(*Y).c1;
         (*X).c2+=c*(*Y).c2;
         (*X).c3+=c*(*Y).c3;         
         (*X).c4+=c*(*Y).c4;
         (*X).c5+=c*(*Y).c5;
         (*X).c6+=c*(*Y).c6;         
         (*X).c7+=c*(*Y).c7;
         (*X).c8+=c*(*Y).c8;
         (*X).c9+=c*(*Y).c9;

         Y+=1;
      }
   }
}


static int set_Xfld(int n,int k,int l,int idet)
{
   int nb,isw,ifail;
   double c;
   spinor_dble *r,*s,*rm;
   u3_alg_dble *X,*Xm;
   block_t *b;
   hmc_t *hflds;
   lat_parms_t lat;

   b=blk_list(HMC_BLOCKS,&nb,&isw)+n;   
   hflds=hmcflds()+n;

   ifail=0;
   lat=lat_parms();
   c=lat.csw/16.0;
   
   X=Xfld;
   r=(*b).sd[k];
   s=(*b).sd[l];
   rm=r+(*b).vol;

   for (;r<rm;r++)
   {
      vec2u3alg(r,s,X);
      Xm=X+6;
      
      for (;X<Xm;X++)
      {         
         (*X).c1*=c;
         (*X).c2*=c;
         (*X).c3*=c;         
         (*X).c4*=c;
         (*X).c5*=c;
         (*X).c6*=c;         
         (*X).c7*=c;
         (*X).c8*=c;
         (*X).c9*=c;         
      }

      s+=1;
   }

   if (idet==1)
   {
      if (query_blk_flags(HMC_BLOCKS,SWD_UP2DATE,n)!=1)
         blk_sw_term(n);

      if (query_blk_flags(HMC_BLOCKS,SWD_E_INVERTED,n)!=1)
         ifail|=invert_blk_swd(HMC_BLOCKS,n,EVEN_PTS);

      if (query_blk_flags(HMC_BLOCKS,SWD_O_INVERTED,n)!=1)
         ifail|=invert_blk_swd(HMC_BLOCKS,n,ODD_PTS);

      det2X((*b).vol,-2.0*c,(*b).swd,Xfld);
   }

   return ifail;
}


static void acc_sw_frc(su3_dble *ub,hmc_t *hflds)
{
   int *pt,*pl,*plm,mu,nu,ipln;
   su3_alg_dble *fb,*fr[7],*frr;
   u3_alg_dble *X[6],*Xb;

   fb=(*hflds).frc;
   pt=(*hflds).pts;
   pl=(*hflds).plq;
   plm=pl+14*(*hflds).nplq;
   ipln=0;

   for (;pl<plm;pl+=14)
   {
      mu=(pl[0]&0x7)>>1;
      nu=(pl[2]&0x7)>>1;
      if (mu==0)
         ipln=-1;
      else if (mu==1)
         ipln=1;
      else 
         ipln=2;

      Xb=Xfld+ipln+nu;
      X[0]=Xb+6*pt[0];
      X[1]=Xb+6*pt[1];
      X[2]=Xb+6*pt[2];
      X[3]=Xb+6*pt[3];      
      X[4]=Xb+6*pt[4];
      X[5]=Xb+6*pt[5];
      pt+=6;

#if (defined SSE2)
      _prefetch_su3(X[0]);
      _prefetch_su3(X[1]);
#endif

      fr[0]=fb+pl[0];
      fr[1]=fb+pl[1];
      fr[2]=fb+pl[2];
      fr[3]=fb+pl[3];
      fr[4]=fb+pl[4];
      fr[5]=fb+pl[5];
      fr[6]=fb+pl[6];

#if (defined SSE2)
      _prefetch_su3_alg_dble(fr[0]);
#endif      
      
      if ((pl[7]+pl[8]+pl[9]+pl[10])!=0)
      {
#if (defined SSE2)
         _prefetch_su3(X[2]);
         _prefetch_su3(X[3]);
#endif

         su3xsu3(ub+pl[0],ub+pl[1],&w1);

#if (defined SSE2)
         _prefetch_su3_alg_dble(fr[1]);
         _prefetch_su3_alg_dble(fr[2]);
         _prefetch_su3_alg_dble(fr[3]);
#endif      

         su3dagxsu3dag(ub+pl[3],ub+pl[2],&w2);

         if ((pl[8]+pl[9]+pl[10])!=0)
         {
            u3algxsu3(X[0],&w1,&w3);

            if (pl[9]!=0)
            {
               prod2su3alg(&w3,&w2,&Yw);
               frr=fr[2];
               _su3_alg_add_assign(*frr,Yw);
            }

            if ((pl[8]+pl[10])!=0)
            {
               prod2su3alg(&w2,&w3,&Yw);

               if (pl[8]!=0)
               {
                  frr=fr[1];
                  _su3_alg_sub_assign(*frr,Yw);
               }

               if (pl[10]!=0)
               {
                  frr=fr[3];
                  _su3_alg_add_assign(*frr,Yw);
               }
            }
         }

         if ((pl[7]+pl[9]+pl[10])!=0)
         {
            su3xu3alg(&w1,X[2],&w3);

            if (pl[10]!=0)
            {
               prod2su3alg(&w2,&w3,&Yw);
               frr=fr[3];
               _su3_alg_add_assign(*frr,Yw);
            }

            if ((pl[7]+pl[9])!=0)
            {
               prod2su3alg(&w3,&w2,&Yw);

               if (pl[7]!=0)
               {
                  frr=fr[0];
                  _su3_alg_sub_assign(*frr,Yw);
               }

               if (pl[9]!=0)
               {
                  frr=fr[2];
                  _su3_alg_add_assign(*frr,Yw);
               }
            }
         }

         if (pl[8]!=0)
         {
            u3algxsu3(X[2],&w2,&w3);
            prod2su3alg(&w3,&w1,&Yw);
            frr=fr[1];
            _su3_alg_sub_assign(*frr,Yw);
         }

         if (pl[7]!=0)
         {
            su3xu3alg(&w2,X[0],&w3);
            prod2su3alg(&w1,&w3,&Yw);
            frr=fr[0];
            _su3_alg_sub_assign(*frr,Yw);
         }
         
         u3algxsu3(X[1],ub+pl[1],&w3);
         su3xsu3(ub+pl[0],&w3,&w3);

         if ((pl[7]+pl[9])!=0)
         {
            prod2su3alg(&w3,&w2,&Yw);

            if (pl[7]!=0)
            {
               frr=fr[0];
               _su3_alg_sub_assign(*frr,Yw);
            }

            if (pl[9]!=0)
            {
               frr=fr[2];
               _su3_alg_add_assign(*frr,Yw);
            }
         }

         if ((pl[8]+pl[10])!=0)
         {
            prod2su3alg(&w2,&w3,&Yw);

            if (pl[8]!=0)
            {
               frr=fr[1];
               _su3_alg_sub_assign(*frr,Yw);
            }

            if (pl[10]!=0)
            {
               frr=fr[3];
               _su3_alg_add_assign(*frr,Yw);
            }
         }

         u3algxsu3dag(X[3],ub+pl[2],&w3);
         su3dagxsu3(ub+pl[3],&w3,&w3);

         if ((pl[7]+pl[9])!=0)
         {
            prod2su3alg(&w1,&w3,&Yw);

            if (pl[7]!=0)
            {
               frr=fr[0];
               _su3_alg_sub_assign(*frr,Yw);
            }

            if (pl[9]!=0)
            {
               frr=fr[2];
               _su3_alg_add_assign(*frr,Yw);
            }
         }

         if ((pl[8]+pl[10])!=0)
         {
            prod2su3alg(&w3,&w1,&Yw);

            if (pl[8]!=0)
            {
               frr=fr[1];
               _su3_alg_sub_assign(*frr,Yw);
            }

            if (pl[10]!=0)
            {
               frr=fr[3];
               _su3_alg_add_assign(*frr,Yw);
            }
         }
      }
      
      if ((pl[7]+pl[11]+pl[12]+pl[13])!=0)
      {
#if (defined SSE2)
         _prefetch_su3(X[4]);
         _prefetch_su3(X[5]);
#endif      

         su3xsu3dag(ub+pl[0],ub+pl[4],&w1);

#if (defined SSE2)
         _prefetch_su3_alg_dble(fr[4]);
         _prefetch_su3_alg_dble(fr[5]);
         _prefetch_su3_alg_dble(fr[6]);         
#endif      
         
         su3dagxsu3(ub+pl[6],ub+pl[5],&w2);

         if ((pl[11]+pl[12]+pl[13])!=0)
         {
            u3algxsu3(X[0],&w1,&w3);

            if (pl[12]!=0)
            {
               prod2su3alg(&w3,&w2,&Yw);
               frr=fr[5];
               _su3_alg_add_assign(*frr,Yw);
            }

            if ((pl[11]+pl[13])!=0)
            {
               prod2su3alg(&w2,&w3,&Yw);

               if (pl[11]!=0)
               {
                  frr=fr[4];
                  _su3_alg_sub_assign(*frr,Yw);
               }

               if (pl[13]!=0)
               {
                  frr=fr[6];
                  _su3_alg_sub_assign(*frr,Yw);
               }
            }
         }

         if ((pl[7]+pl[12]+pl[13])!=0)
         {
            su3xu3alg(&w1,X[4],&w3);

            if (pl[13]!=0)
            {
               prod2su3alg(&w2,&w3,&Yw);
               frr=fr[6];
               _su3_alg_sub_assign(*frr,Yw);
            }

            if ((pl[7]+pl[12])!=0)
            {
               prod2su3alg(&w3,&w2,&Yw);

               if (pl[7]!=0)
               {
                  frr=fr[0];
                  _su3_alg_add_assign(*frr,Yw);
               }

               if (pl[12]!=0)
               {
                  frr=fr[5];
                  _su3_alg_add_assign(*frr,Yw);
               }
            }
         }

         if (pl[11]!=0)
         {
            u3algxsu3(X[4],&w2,&w3);
            prod2su3alg(&w3,&w1,&Yw);
            frr=fr[4];
            _su3_alg_sub_assign(*frr,Yw);
         }

         if (pl[7]!=0)
         {
            su3xu3alg(&w2,X[0],&w3);
            prod2su3alg(&w1,&w3,&Yw);
            frr=fr[0];
            _su3_alg_add_assign(*frr,Yw);
         }
         
         u3algxsu3dag(X[1],ub+pl[4],&w3);
         su3xsu3(ub+pl[0],&w3,&w3);

         if ((pl[7]+pl[12])!=0)
         {
            prod2su3alg(&w3,&w2,&Yw);

            if (pl[7]!=0)
            {
               frr=fr[0];
               _su3_alg_add_assign(*frr,Yw);
            }

            if (pl[12]!=0)
            {
               frr=fr[5];
               _su3_alg_add_assign(*frr,Yw);
            }
         }

         if ((pl[11]+pl[13])!=0)
         {
            prod2su3alg(&w2,&w3,&Yw);

            if (pl[11]!=0)
            {
               frr=fr[4];
               _su3_alg_sub_assign(*frr,Yw);
            }

            if (pl[13]!=0)
            {
               frr=fr[6];
               _su3_alg_sub_assign(*frr,Yw);
            }
         }

         u3algxsu3(X[5],ub+pl[5],&w3);
         su3dagxsu3(ub+pl[6],&w3,&w3);

         if ((pl[7]+pl[12])!=0)
         {
            prod2su3alg(&w1,&w3,&Yw);

            if (pl[7]!=0)
            {
               frr=fr[0];
               _su3_alg_add_assign(*frr,Yw);
            }

            if (pl[12]!=0)
            {
               frr=fr[5];
               _su3_alg_add_assign(*frr,Yw);
            }
         }

         if ((pl[11]+pl[13])!=0)
         {
            prod2su3alg(&w3,&w1,&Yw);

            if (pl[11]!=0)
            {
               frr=fr[4];   
               _su3_alg_sub_assign(*frr,Yw);
            }

            if (pl[13]!=0)
            {
               frr=fr[6];
               _su3_alg_sub_assign(*frr,Yw);
            }
         }
      }
   }
}


static void cnv_sw_frc(su3_dble *ub,hmc_t *hflds)
{
   int ilk,*lk,*lkm;
   su3_alg_dble *fb,*fr,*frn;
   su3_dble *u;

   fb=(*hflds).frc;
   frn=fb;
   lk=(*hflds).lks;
   lkm=lk+(*hflds).nlks;
   
   for (;lk<lkm;lk++)
   {
      ilk=(*lk);
      u=ub+ilk;
      fr=fb+ilk;
      
      if ((ilk&0x1)!=0)
         rotate_su3alg(u,fr);

      *frn=*fr;
      frn+=1;
   }
}


int sw_frc(int n,int k,int l,int idet)
{
   int nb,isw,ifail;
   su3_dble *ub;
   block_t *b;
   hmc_t *hflds;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   if (error_loc((b==NULL)||(hflds==NULL),1,"sw_frc [genfrc.c]",
                 "Block grid or HMC fields are not allocated"))
      return 2;

   if (error_loc((n<0)||(n>=nb)||(k<0)||(k>=(*b).nsd)||(l<0)||(l>=(*b).nsd),1,
                 "sw_frc [genfrc.c]","Parameters are out of range"))
      return 3;

   if ((*b).vol>size)
   {
      if (size>0)
         afree(Xfld);
         
      Xfld=amalloc(6*(*b).vol*sizeof(u3_alg_dble),ALIGN);
      if (error_loc(Xfld==NULL,1,"sw_frc [genfrc.c]",
                    "Unable to allocate workspace"))
      {
         size=0;
         return 4;
      }

      size=(*b).vol;
   }
   
   ifail=set_Xfld(n,k,l,idet);

   b+=n;
   hflds+=n;
   ub=(*b).ud[0];
   set_frc2zero(hflds);
   acc_sw_frc(ub,hflds);
   cnv_sw_frc(ub,hflds);

   return ifail;
}


void hop_frc(int n,int k,int l)
{
   int nb,isw;
   int ix,iy,is,mu;
   int *lk,*lkm;
   su3_dble *ub;
   su3_alg_dble *fr;
   spinor_dble *rb,*sb;
   block_t *b;
   hmc_t *hflds;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   hflds=hmcflds();

   if (error_loc((b==NULL)||(hflds==NULL),1,"hop_frc [genfrc.c]",
                 "Block grid or HMC fields are not allocated"))
      return;

   if (error_loc((n<0)||(n>=nb)||(k<0)||(k>=(*b).nsd)||(l<0)||(l>=(*b).nsd),1,
                 "hop_frc [genfrc.c]","Parameters are out of range"))
      return;
   
   b+=n;
   hflds+=n;
   rb=(*b).sd[k];
   sb=(*b).sd[l];   
   ub=(*b).ud[0];
   fr=(*hflds).frc;
   lk=(*hflds).lks;
   lkm=lk+(*hflds).nlks;
   
   for (;lk<lkm;lk++)
   {
      is=(*lk)&0x1;
      mu=((*lk)>>1)&0x3;
      ix=((*lk)>>3)+(*b).volh;
      
      if (is==0)
         iy=(*b).iup[ix][mu];
      else
      {
         iy=ix;
         ix=(*b).idn[iy][mu];
      }

      spin2mat[mu](rb+ix,rb+iy,sb+ix,sb+iy,&w1);
      prod2su3alg(ub+(*lk),&w1,&Yw);

      (*fr).c1+=0.5*Yw.c1;
      (*fr).c2+=0.5*Yw.c2;
      (*fr).c3+=0.5*Yw.c3;
      (*fr).c4+=0.5*Yw.c4;
      (*fr).c5+=0.5*Yw.c5;
      (*fr).c6+=0.5*Yw.c6;
      (*fr).c7+=0.5*Yw.c7;
      (*fr).c8+=0.5*Yw.c8;
      fr+=1;
   }
}

