
/*******************************************************************************
*
* File hmcflds.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Allocation of the HMC auxiliary fields
*
* The externally accessible functions are
*
*   void alloc_hmcflds(void)
*     Allocates the auxiliary fields (index arrays, momenta and pseudo-fermion
*     fields) that are required for the HMC algorithm
*
*   void free_hmcflds(void)
*     Frees the fields previously allocated by alloc_hmcflds()
*
*   hmc_t *hmcflds(void)
*     Returns the pointer to the array hflds[0],..,hflds[nb-1] of hmc_t
*     structures that contain the HMC auxiliary fields associated to the
*     blocks b[0],..,b[nb-1] in the HMC_BLOCKS grid. A NULL pointer is
*     returned if the fields are not allocated
*
* Notes:
*
* The programs alloc_hmcflds and free_hmcflds are called by the programs
* that allocate and free the HMC_BLOCKS grid. It is therefore unnecessary
* and not recommended to call them from anywhere else. The program hmcflds
* does not involve any communications and may be called locally
*
* The hmc_t structure contains the pointers to the various fields on a given
* block which are needed for the HMC algorithm. The corresponding type is
* declared in the header file hmc.h
*
* The force field and the index arrays lks,plq,pts are shared between the
* blocks. All other fields are not shared, but imb,ifc and chi are only
* allocated on the black blocks
*
*******************************************************************************/

#define HMCFLDS_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "block.h"
#include "hmc.h"
#include "global.h"

static hmc_t *hflds=NULL;


static int is_on_bnd(block_t *b,int ix)
{
   int mu;

   if (ix>=(*b).vol)
      return 1;
   
   for (mu=0;mu<4;mu++)
      if (((*b).iup[ix][mu]>=(*b).vol)||((*b).idn[ix][mu]>=(*b).vol))
         return 1;

   return 0;
}


static void set_lks(block_t *b,int *lks)
{
   int n,ix,iy,mu;

   n=0;
   
   for (ix=(*b).volh;ix<(*b).vol;ix++)
   {
      for (mu=0;mu<4;mu++)
      {
         iy=(*b).iup[ix][mu];

         if (iy<(*b).vol)
         {
            if ((is_on_bnd(b,ix)+is_on_bnd(b,iy))!=2)
            {
               lks[n]=8*(ix-(*b).volh)+2*mu;
               n+=1;
            }
         }

         iy=(*b).idn[ix][mu];

         if (iy<(*b).vol)
         {
            if ((is_on_bnd(b,ix)+is_on_bnd(b,iy))!=2)
            {         
               lks[n]=8*(ix-(*b).volh)+2*mu+1;
               n+=1;
            }
         }
      }
   }
}


static void set_plq(block_t *b,int *plq,int *pts)
{
   int i,mu,nu;
   int ix,iy,iz,iw,ir,is;
   int ibx,iby,ibz,ibw,ibr,ibs;

   for (ix=(*b).volh;ix<(*b).vol;ix++)
   {
      ibx=is_on_bnd(b,ix);
      
      for (mu=0;mu<3;mu++)
      {
         iy=(*b).iup[ix][mu];
               
         if (iy<(*b).vol)
         {
            iby=is_on_bnd(b,iy);
         
            for (nu=(mu+1);nu<4;nu++)
            {
               iz=(*b).iup[iy][nu];
               iw=(*b).iup[ix][nu];
               ir=(*b).idn[iy][nu];
               is=(*b).idn[ix][nu];

               ibz=is_on_bnd(b,iz);
               ibw=is_on_bnd(b,iw);
               ibr=is_on_bnd(b,ir);
               ibs=is_on_bnd(b,is);

               if ((ibx+iby+ibz+ibw+ibr+ibs)!=6)
               {
                  i=8*(ix-(*b).volh);
                  plq[0]=i+2*mu;
                  plq[1]=i;
                  plq[2]=i+2*nu;
                  plq[3]=i;
                  plq[4]=i;
                  plq[5]=i+2*nu+1;
                  plq[6]=i;

                  for (i=7;i<14;i++)
                     plq[i]=0;
               
                  if (iz<(*b).vol)
                  {
                     plq[1]=8*(iz-(*b).volh)+2*nu+1;
                     plq[3]=8*(iz-(*b).volh)+2*mu+1;

                     if ((ibx+iby)!=2)
                        plq[7]=1;
                     if ((iby+ibz)!=2)
                        plq[8]=1;
                     if ((ibx+ibw)!=2)
                        plq[9]=1;
                     if ((ibz+ibw)!=2)
                        plq[10]=1;
                  }

                  if (ir<(*b).vol)
                  {
                     plq[4]=8*(ir-(*b).volh)+2*nu;
                     plq[6]=8*(ir-(*b).volh)+2*mu+1;

                     if ((ibr+iby)!=2)
                        plq[11]=1;
                     if ((ibx+ibs)!=2)
                        plq[12]=1;
                     if ((ibr+ibs)!=2)
                        plq[13]=1;
                  }

                  pts[0]=ix;
                  pts[1]=iy;
                  pts[2]=iz;
                  pts[3]=iw;
                  pts[4]=ir;
                  pts[5]=is;
                  
                  plq+=14;
                  pts+=6;
               }
            }
         }
      }
   }
}


static void set_imb(block_t *b,int *imb,int *ifc)
{
   int ix,mu,ib;

   for (ix=0;ix<(*b).vol;ix++)
   {
      ib=0;

      for (mu=0;mu<4;mu++)
      {
         if ((*b).iup[ix][mu]>=(*b).vol)
         {
            ib+=1;
            *ifc=2*mu+1;
         }

         if ((*b).idn[ix][mu]>=(*b).vol)
         {
            ib+=1;
            *ifc=2*mu;
         }         
      }
      
      if (ib>=1)
      {
         *imb=(*b).imb[ix];

         if (ib!=1)
            *ifc=8;

         imb+=1;
         ifc+=1;
      }
   }
}


static void alloc_blkflds(block_t *b,hmc_t *bflds,int ifr,int ich)
{
   int nlks,nplq,nfrc,nbnd;
   int *lks,*plq,*pts,*imb,*ifc,*bs;
   su3_alg_dble *mom,*frc;
   spinor_dble *phi,*chi;

   bs=(*b).bs;
   
   nlks=((bs[0]-1)*(bs[1]-2)*(bs[2]-2)*(bs[3]-2)+
         (bs[0]-2)*(bs[1]-1)*(bs[2]-2)*(bs[3]-2)+
         (bs[0]-2)*(bs[1]-2)*(bs[2]-1)*(bs[3]-2)+
         (bs[0]-2)*(bs[1]-2)*(bs[2]-2)*(bs[3]-1));

   nplq=((bs[0]-1)*(bs[1]  )*(bs[2]-2)*(bs[3]-2)+
         (bs[0]-1)*(bs[1]-2)*(bs[2]  )*(bs[3]-2)+
         (bs[0]-1)*(bs[1]-2)*(bs[2]-2)*(bs[3]  )+
         (bs[0]-2)*(bs[1]-1)*(bs[2]  )*(bs[3]-2)+
         (bs[0]-2)*(bs[1]-1)*(bs[2]-2)*(bs[3]  )+
         (bs[0]-2)*(bs[1]-2)*(bs[2]-1)*(bs[3]  ))/2;

   nfrc=8*(*b).volh;
   nbnd=(*b).vol-(bs[0]-2)*(bs[1]-2)*(bs[2]-2)*(bs[3]-2);

   error((nlks<=0)||(nplq<=0)||(nfrc<=0)||(nbnd<=0),1,
         "alloc_blkflds [hmcflds.c]",
         "Block is too small or not properly initialized");

   if ((ifr!=0)||(ich!=0))
      lks=amalloc((ifr*(nlks+20*nplq)+ich*2*nbnd)*sizeof(int),3);
   else
      lks=NULL;
   mom=amalloc((nlks+ifr*nfrc)*sizeof(su3_alg_dble),ALIGN);
   phi=amalloc(((*b).volh+ich*nbnd)*sizeof(spinor_dble),ALIGN);
   
   error((((ifr!=0)||(ich!=0))&&(lks==NULL))||(mom==NULL)||(phi==NULL),1,
         "alloc_blkflds [hmcflds.c]","Unable to allocate field arrays");

   if (ifr==1)
   {
      plq=lks+nlks;
      pts=plq+14*nplq;
      imb=pts+6*nplq;
      frc=mom+nlks;
      set_lks(b,lks);
      set_plq(b,plq,pts);
   }
   else
   {
      imb=lks;
      lks=NULL;
      plq=NULL;
      pts=NULL;
      frc=NULL;
   }
   
   if (ich==1)
   {
      ifc=imb+nbnd;
      chi=phi+(*b).volh;
      set_imb(b,imb,ifc);
   }
   else
   {
      imb=NULL;
      ifc=NULL;
      chi=NULL;
   }
   
   set_alg2zero(nlks+ifr*nfrc,mom);
   set_sd2zero((*b).volh+ich*nbnd,phi);
   
   (*bflds).nlks=nlks;
   (*bflds).nplq=nplq;   
   (*bflds).nbnd=nbnd;

   (*bflds).lks=lks;
   (*bflds).plq=plq;
   (*bflds).pts=pts;
   (*bflds).imb=imb;
   (*bflds).ifc=ifc;      
   (*bflds).mom=mom;
   (*bflds).frc=frc;
   (*bflds).phi=phi;
   (*bflds).chi=chi;   
}


void alloc_hmcflds(void)
{
   int n,nb,isw;
   block_t *b0,*b,*bm;
   hmc_t *bflds;

   if (hflds!=NULL)
      return;
   
   b0=blk_list(HMC_BLOCKS,&nb,&isw);
   error(b0==NULL,1,"alloc_hmcflds [hmcflds.c]",
         "HMC block grid is not allocated");
   
   hflds=amalloc(nb*sizeof(hmc_t),3);
   error(hflds==NULL,1,"alloc_hmcflds [hmcflds.c]",
         "Unable to allocate HMC fields");

   if (isw==0)
      n=0;
   else
      n=(nb/2);

   b=b0+n;
   bflds=hflds+n;
   bm=b+(nb/2);
   
   alloc_blkflds(b,bflds,1,1);
   b+=1;
   bflds+=1;
      
   for (;b<bm;b++)
   {
      alloc_blkflds(b,bflds,0,1);

      (*bflds).lks=hflds[n].lks;
      (*bflds).plq=hflds[n].plq;
      (*bflds).pts=hflds[n].pts;
      (*bflds).frc=hflds[n].frc;
      bflds+=1;
   }

   if (isw==1)
   {
      b-=nb;
      bflds-=nb;
   }

   bm=b+(nb/2);

   for (;b<bm;b++)
   {
      alloc_blkflds(b,bflds,0,0);

      (*bflds).lks=hflds[n].lks;
      (*bflds).plq=hflds[n].plq;
      (*bflds).pts=hflds[n].pts;
      (*bflds).frc=hflds[n].frc;
      bflds+=1;
   }      
}


void free_hmcflds(void)
{
   int n,nb,isw;
   block_t *b0,*b,*bm;
   hmc_t *bflds;

   if (hflds==NULL)
      return;

   b0=blk_list(HMC_BLOCKS,&nb,&isw);
   
   if (isw==0)
      n=0;
   else
      n=(nb/2);

   b=b0+n;
   bflds=hflds+n;
   bm=b+(nb/2);

   afree((*bflds).lks);
   afree((*bflds).mom);   
   afree((*bflds).phi);   

   b+=1;
   bflds+=1;
      
   for (;b<bm;b++)
   {
      afree((*bflds).imb);
      afree((*bflds).mom);   
      afree((*bflds).phi); 
      bflds+=1;
   }

   if (isw==1)
   {
      b-=nb;
      bflds-=nb;
   }

   bm=b+(nb/2);

   for (;b<bm;b++)
   {
      afree((*bflds).mom);   
      afree((*bflds).phi); 
      bflds+=1;
   }      
   
   afree(hflds);
   hflds=NULL;
}


hmc_t *hmcflds(void)
{
   return hflds;
}
