
/*******************************************************************************
*
* File bicg.c
*
* Copyright (C) 2005 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generic BICGstab inversion program for operators Op acting on
* double-precision spinor fields
*
* The externally accessible function is
*
*   double bicg(int vol,int icom,void (*Op)(spinor_dble *pk,spinor_dble *pl),
*               spinor_dble *wsd[],int nmx,double res,spinor_dble *pk,
*               spinor_dble *pl,int *status)
*     Solution of the equation Op*pl[]=pk[] for given pk[]. The parameters
*     vol, nmx and res denote the number of spinors in each field, the
*     maximal number of BICGstab iterations and the desired reduction of
*     the norm of the residue. Scalar products etc. are globally summed if
*     (and only if) icom=1. A workspace of 5 double-precision fields is
*     needed that are pointed to by the array wsd[]. On exit status reports
*     the number of BICGstab iterations that were required, or a negative
*     value if the program failed. In all cases the norm of the last residue
*     is returned
*
* Notes:
*
* The program for the operator must be externally defined and is assumed to
* have the following properties:
*
*   void Op(spinor_dble *pk,spinor_dble *pl)
*     Application of the operator Op to the double-precision field pk[]
*     and assignement of the result to pl[]. On exit the field pk[] is
*     unchanged
*
*******************************************************************************/

#define BICG_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "linalg.h"
#include "global.h"

static int nsp,com,my_rank;
static complex_dble ai,bi,oi,ri;
static spinor_dble *pdb,*pdx,*pdp,*pdr,*pdap,*pdas,*pdr0;

#if (defined SSE2)
#include "sse2.h"

static void update_p(void)
{
   su3_vector_dble *rdp,*rdr,*rdm;

   _sse_load_cmplx_dble(bi);

   rdr=(su3_vector_dble*)(pdr);
   rdm=(su3_vector_dble*)(pdp+nsp);

   for (rdp=(su3_vector_dble*)(pdp);rdp<rdm;rdp+=4)
   {
      _prefetch_spinor_dble(rdp+4);
      _sse_load_dble(*(rdp  ));
      _sse_mulc_dble();
      _sse_load_dble(*(rdr  ));
      _sse_vector_add_dble();
      _sse_store_dble(*(rdp  ));

      _sse_load_dble(*(rdp+1));
      _sse_mulc_dble();
      _sse_load_dble(*(rdr+1));
      _sse_vector_add_dble();
      _sse_store_dble(*(rdp+1));

      _prefetch_spinor_dble(rdr+4);
      _sse_load_dble(*(rdp+2));
      _sse_mulc_dble();
      _sse_load_dble(*(rdr+2));
      _sse_vector_add_dble();
      _sse_store_dble(*(rdp+2));

      _sse_load_dble(*(rdp+3));
      _sse_mulc_dble();
      _sse_load_dble(*(rdr+3));
      _sse_vector_add_dble();
      _sse_store_dble(*(rdp+3));

      rdr+=4;
   }
}

#else

static void update_p(void)
{
   complex_dble z;
   complex_dble *rdp,*rdr,*rdm;

   rdr=(complex_dble*)(pdr);
   rdm=(complex_dble*)(pdp+nsp);

   for (rdp=(complex_dble*)(pdp);rdp<rdm;rdp+=6)
   {
      z.re=(*(rdp  )).re;
      z.im=(*(rdp  )).im;
      (*(rdp  )).re=(*(rdr  )).re+bi.re*z.re-bi.im*z.im;
      (*(rdp  )).im=(*(rdr  )).im+bi.re*z.im+bi.im*z.re;

      z.re=(*(rdp+1)).re;
      z.im=(*(rdp+1)).im;
      (*(rdp+1)).re=(*(rdr+1)).re+bi.re*z.re-bi.im*z.im;
      (*(rdp+1)).im=(*(rdr+1)).im+bi.re*z.im+bi.im*z.re;

      z.re=(*(rdp+2)).re;
      z.im=(*(rdp+2)).im;
      (*(rdp+2)).re=(*(rdr+2)).re+bi.re*z.re-bi.im*z.im;
      (*(rdp+2)).im=(*(rdr+2)).im+bi.re*z.im+bi.im*z.re;

      z.re=(*(rdp+3)).re;
      z.im=(*(rdp+3)).im;
      (*(rdp+3)).re=(*(rdr+3)).re+bi.re*z.re-bi.im*z.im;
      (*(rdp+3)).im=(*(rdr+3)).im+bi.re*z.im+bi.im*z.re;

      z.re=(*(rdp+4)).re;
      z.im=(*(rdp+4)).im;
      (*(rdp+4)).re=(*(rdr+4)).re+bi.re*z.re-bi.im*z.im;
      (*(rdp+4)).im=(*(rdr+4)).im+bi.re*z.im+bi.im*z.re;

      z.re=(*(rdp+5)).re;
      z.im=(*(rdp+5)).im;
      (*(rdp+5)).re=(*(rdr+5)).re+bi.re*z.re-bi.im*z.im;
      (*(rdp+5)).im=(*(rdr+5)).im+bi.re*z.im+bi.im*z.re;

      rdr+=6;
   }
}

#endif

static void bicg_init(int vol,int icom,spinor_dble *wsd[],
                      spinor_dble *pk,spinor_dble *pl)
{
   nsp=vol;
   com=icom;
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   pdb=pk;
   pdx=pl;

   pdp=wsd[0];
   pdr=wsd[1];
   pdap=wsd[2];
   pdas=wsd[3];
   pdr0=wsd[4];

   set_sd2zero(nsp,pdx);
   assign_sd2sd(nsp,pdb,pdp);
   assign_sd2sd(nsp,pdb,pdr);
   assign_sd2sd(nsp,pdb,pdr0);

   ri.re=norm_square_dble(nsp,com,pdb);
   ri.im=0.0;
}


static void bicg_step(void (*Op)(spinor_dble *pk,spinor_dble *pl))
{
   double r;
   complex_dble z,w;

   (*Op)(pdp,pdap);
   z=spinor_prod_dble(nsp,com,pdr0,pdap);
   r=z.re*z.re+z.im*z.im;
   ai.re=(ri.re*z.re+ri.im*z.im)/r;
   ai.im=(ri.im*z.re-ri.re*z.im)/r;

   z.re=-ai.re;
   z.im=-ai.im;
   mulc_spinor_add_dble(nsp,pdr,pdap,z);

   (*Op)(pdr,pdas);
   z=spinor_prod_dble(nsp,com,pdas,pdr);
   r=norm_square_dble(nsp,com,pdas);
   oi.re=z.re/r;
   oi.im=z.im/r;

   mulc_spinor_add_dble(nsp,pdx,pdp,ai);
   mulc_spinor_add_dble(nsp,pdx,pdr,oi);

   z.re=-oi.re;
   z.im=-oi.im;
   mulc_spinor_add_dble(nsp,pdr,pdas,z);

   z.re=ri.re*oi.re-ri.im*oi.im;
   z.im=ri.re*oi.im+ri.im*oi.re;
   ri=spinor_prod_dble(nsp,com,pdr0,pdr);
   w.re=ri.re*ai.re-ri.im*ai.im;
   w.im=ri.re*ai.im+ri.im*ai.re;
   r=z.re*z.re+z.im*z.im;
   bi.re=(w.re*z.re+w.im*z.im)/r;
   bi.im=(w.im*z.re-w.re*z.im)/r;

   z.re=-oi.re;
   z.im=-oi.im;
   mulc_spinor_add_dble(nsp,pdp,pdap,z);
   update_p();
}


double bicg(int vol,int icom,void (*Op)(spinor_dble *pk,spinor_dble *pl),
            spinor_dble *wsd[],int nmx,double res,spinor_dble *pk,
            spinor_dble *pl,int *status)
{
   int ncg,iprms[2];
   double xn,rn,tol,dprms[1];

   if ((icom==1)&&(NPROC>1))
   {
      iprms[0]=vol;
      iprms[1]=nmx;
      dprms[0]=res;

      MPI_Bcast(iprms,2,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(dprms,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

      error((iprms[0]!=vol)||(iprms[1]!=nmx)||(dprms[0]!=res),1,
            "bicg [bicg.c]","Parameters are not global");

      error_root((vol<=0)||(nmx<1)||(res<=DBL_EPSILON),1,
                 "bicg [bicg.c]","Improper choice of vol,nmx or res");
   }
   else if (error_loc((vol<=0)||(nmx<1)||(res<=DBL_EPSILON),1,
                      "bicg [bicg.c]","Improper choice of vol,nmx or res"))
   {
      (*status)=0;
      return 1.0;
   }

   bicg_init(vol,icom,wsd,pk,pl);

   ncg=0;
   rn=sqrt(ri.re);
   tol=res*rn;

   while (rn>tol)
   {
      bicg_step(Op);
      ncg+=1;

      xn=norm_square_dble(nsp,com,pdx);
      xn=sqrt(xn);
      rn=norm_square_dble(nsp,com,pdr);
      rn=sqrt(rn);

      /* if ((my_rank==0)&((ncg%10)==0))
         printf("ncg=%d, xn=%.2e, rn=%.2e\n",ncg,xn,rn); */

      if ((ncg>=nmx)&&(rn>tol))
      {
         (*status)=-1;
         return rn;
      }

      if ((10.0*DBL_EPSILON*xn)>tol)
      {
         (*status)=-2;
         return rn;
      }
   }

   (*status)=ncg;
   return rn;
}

