
/*******************************************************************************
*
* File cgne.c
*
* Copyright (C) 2005, 2008 Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Generic CG inversion program for operators of the form Op*Op where Op is
* any hermitian operator that acts on spinor fields
*
* The externally accessible function is
*
*   double cgne(int vol,int icom,void (*Op)(spinor *pk,spinor *pl),
*               void (*Op_dble)(spinor_dble *pk,spinor_dble *pl),
*               spinor *ws[],spinor_dble *wsd[],int nmx,double res,
*               spinor_dble *pk,spinor_dble *pl,int *status)
*     Solution of the equation Op*Op*pl[]=pk[] for given pk[]. The parameters
*     vol, nmx and res denote the number of spinors in each field, the
*     maximal number of CG iterations and the desired reduction of the norm
*     of the residue. Scalar products etc. are globally summed if (and only
*     if) icom=1. A workspace of 5 single-precision and 2 double-precision
*     fields is needed that are pointed to by the arrays ws[] and wsd[]. On
*     exit status reports the number of conjugate gradient iterations that
*     were required, or a negative value if the program failed. In all cases
*     the norm of the last residue is returned
*
* Notes:
*
* This program uses single-precision arithmetic to reduce the execution
* time but obtains the solution with double-precision accuracy
*
* The programs for the single- and double-precision implementations of the
* operator must be externally defined and are assumed to have the following
* properties:
*
*   void Op(spinor *pk,spinor *pl)
*     Application of the operator Op to the single-precision field pk[]
*     and assignement of the result to pl[]. On exit the field pk[] is
*     unchanged
*
*   void Op_dble(spinor_dble *pk,spinor_dble *pl)
*     Application of the operator Op to the double-precision field pk[]
*     and assignement of the result to pl[]. On exit the field pk[] is
*     unchanged
*
*******************************************************************************/

#define CGNE_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <float.h>
#include "mpi.h"
#include "su3.h"
#include "start.h"
#include "linalg.h"
#include "global.h"

#define PRECISION_LIMIT ((double)(100.0f*FLT_EPSILON))

static int nsp,com,my_rank;
static float rsq,rsq_old,ai,bi;
static spinor *psx,*psr,*psp,*psap,*psw;
static spinor_dble *pdb,*pdx,*pdw,*pdv;

#if (defined SSE)
#include "sse.h"


static void update_g(void)
{
   spinor *rsr,*rsap,*rsm;

   __asm__ __volatile__ ("movss %0, %%xmm6 \n\t"
                         "shufps $0x0, %%xmm6, %%xmm6 \n\t"
                         "mulps %1, %%xmm6"
                         :
                         :
                         "m" (ai),
                         "m" (_sse_sgn)
                         :
                         "xmm6");

   rsap=psap;
   rsr=psr;
   rsm=psr+nsp;

   for (;rsr<rsm;rsr++)
   {
      _sse_spinor_load(*rsap);

      rsap+=4;
      _prefetch_spinor(rsap);
      rsap-=4;      
      
      __asm__ __volatile__ ("mulps %%xmm6, %%xmm0 \n\t"
                            "mulps %%xmm6, %%xmm1 \n\t"
                            "mulps %%xmm6, %%xmm2 \n\t"
                            "mulps %%xmm6, %%xmm3 \n\t"
                            "mulps %%xmm6, %%xmm4 \n\t"
                            "mulps %%xmm6, %%xmm5"
                            :
                            :
                            :
                            "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5");

      rsr+=4;
      _prefetch_spinor(rsr);
      rsr-=4;

      __asm__ __volatile__ ("addps %0, %%xmm0 \n\t"
                            "addps %1, %%xmm1 \n\t"
                            "addps %2, %%xmm2 \n\t"
                            "addps %3, %%xmm3 \n\t"
                            "addps %4, %%xmm4 \n\t"
                            "addps %5, %%xmm5"
                            :
                            :
                            "m" ((*rsr).c1.c1.re),
                            "m" ((*rsr).c1.c3.re),
                            "m" ((*rsr).c2.c2.re),
                            "m" ((*rsr).c3.c1.re),
                            "m" ((*rsr).c3.c3.re),
                            "m" ((*rsr).c4.c2.re)
                            :
                            "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5");

      _sse_spinor_store(*rsr);
      rsap+=1;
   }
}


static void update_xp(void)
{
   spinor *rsx,*rsp,*rsr,*rsm;

   rsp=psp;
   rsr=psr;
   rsx=psx;
   rsm=psx+nsp;

   for (;rsx<rsm;rsx++)
   {
      __asm__ __volatile__ ("movss %0, %%xmm6 \n\t"
                            "shufps $0x0, %%xmm6, %%xmm6"
                            :
                            :
                            "m" (ai)
                            :
                            "xmm6");

      _sse_spinor_load(*rsp);

      rsp+=4;
      _prefetch_spinor(rsp);
      rsp-=4;  
      
      __asm__ __volatile__ ("mulps %%xmm6, %%xmm0 \n\t"
                            "mulps %%xmm6, %%xmm1 \n\t"
                            "mulps %%xmm6, %%xmm2 \n\t"
                            "mulps %%xmm6, %%xmm3 \n\t"
                            "mulps %%xmm6, %%xmm4 \n\t"
                            "mulps %%xmm6, %%xmm5"
                            :
                            :
                            :
                            "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5");

      rsx+=4;
      _prefetch_spinor(rsx);
      rsx-=4;

      __asm__ __volatile__ ("addps %0, %%xmm0 \n\t"
                            "addps %1, %%xmm1 \n\t"
                            "addps %2, %%xmm2 \n\t"
                            "addps %3, %%xmm3 \n\t"
                            "addps %4, %%xmm4 \n\t"
                            "addps %5, %%xmm5"
                            :
                            :
                            "m" ((*rsx).c1.c1.re),
                            "m" ((*rsx).c1.c3.re),
                            "m" ((*rsx).c2.c2.re),
                            "m" ((*rsx).c3.c1.re),
                            "m" ((*rsx).c3.c3.re),
                            "m" ((*rsx).c4.c2.re)
                            :
                            "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5");

      _sse_spinor_store(*rsx); 

      __asm__ __volatile__ ("movss %0, %%xmm6 \n\t"
                            "shufps $0x0, %%xmm6, %%xmm6"
                            :
                            :
                            "m" (bi)
                            :
                            "xmm6");

      _sse_spinor_load(*rsp);

      rsr+=4;
      _prefetch_spinor(rsr);
      rsr-=4;
      
      __asm__ __volatile__ ("mulps %%xmm6, %%xmm0 \n\t"
                            "mulps %%xmm6, %%xmm1 \n\t"
                            "mulps %%xmm6, %%xmm2 \n\t"
                            "mulps %%xmm6, %%xmm3 \n\t"
                            "mulps %%xmm6, %%xmm4 \n\t"
                            "mulps %%xmm6, %%xmm5"
                            :
                            :
                            :
                            "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5");
      
      __asm__ __volatile__ ("addps %0, %%xmm0 \n\t"
                            "addps %1, %%xmm1 \n\t"
                            "addps %2, %%xmm2 \n\t"
                            "addps %3, %%xmm3 \n\t"
                            "addps %4, %%xmm4 \n\t"
                            "addps %5, %%xmm5"
                            :
                            :
                            "m" ((*rsr).c1.c1.re),
                            "m" ((*rsr).c1.c3.re),
                            "m" ((*rsr).c2.c2.re),
                            "m" ((*rsr).c3.c1.re),
                            "m" ((*rsr).c3.c3.re),
                            "m" ((*rsr).c4.c2.re)
                            :
                            "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5");
      
      _sse_spinor_store(*rsp);

      rsp+=1;
      rsr+=1;
   }
}

#else

static void update_g(void)
{
   float *rsr,*rsap,*rsm;

   rsap=(float*)(psap);
   rsm=(float*)(psr+nsp);

   for (rsr=(float*)(psr);rsr<rsm;rsr+=8)
   {
      *(rsr  )-=ai*(*(rsap  ));
      *(rsr+1)-=ai*(*(rsap+1));
      *(rsr+2)-=ai*(*(rsap+2));
      *(rsr+3)-=ai*(*(rsap+3));
      *(rsr+4)-=ai*(*(rsap+4));
      *(rsr+5)-=ai*(*(rsap+5));
      *(rsr+6)-=ai*(*(rsap+6));
      *(rsr+7)-=ai*(*(rsap+7));

      rsap+=8;
   }
}


static void update_xp(void)
{
   float *rsx,*rsr,*rsp,*rsm;

   rsr=(float*)(psr);
   rsp=(float*)(psp);
   rsm=(float*)(psx+nsp);

   for (rsx=(float*)(psx);rsx<rsm;rsx+=8)
   {
      *(rsx  )+=ai*(*(rsp  ));
      *(rsx+1)+=ai*(*(rsp+1));
      *(rsx+2)+=ai*(*(rsp+2));
      *(rsx+3)+=ai*(*(rsp+3));
      *(rsx+4)+=ai*(*(rsp+4));
      *(rsx+5)+=ai*(*(rsp+5));
      *(rsx+6)+=ai*(*(rsp+6));
      *(rsx+7)+=ai*(*(rsp+7));

      *(rsp  )=*(rsr  )+bi*(*(rsp  ));
      *(rsp+1)=*(rsr+1)+bi*(*(rsp+1));
      *(rsp+2)=*(rsr+2)+bi*(*(rsp+2));
      *(rsp+3)=*(rsr+3)+bi*(*(rsp+3));
      *(rsp+4)=*(rsr+4)+bi*(*(rsp+4));
      *(rsp+5)=*(rsr+5)+bi*(*(rsp+5));
      *(rsp+6)=*(rsr+6)+bi*(*(rsp+6));
      *(rsp+7)=*(rsr+7)+bi*(*(rsp+7));

      rsr+=8;
      rsp+=8;
   }
}

#endif

static void cg_init(int vol,int icom,spinor *ws[],spinor_dble *wsd[],
                    spinor_dble *pk,spinor_dble *pl)
{
   nsp=vol;
   com=icom;
   MPI_Comm_rank(MPI_COMM_WORLD,&my_rank);

   psx=ws[0];
   psr=ws[1];
   psp=ws[2];
   psap=ws[3];
   psw=ws[4];

   pdb=pk;
   pdx=pl;
   pdw=wsd[0];
   pdv=wsd[1];

   set_s2zero(nsp,psx);
   assign_sd2s(nsp,pdb,psr);
   assign_s2s(nsp,psr,psp);
   set_sd2zero(nsp,pdx);

   rsq=norm_square(nsp,com,psr);
}


static void cg_step(void (*Op)(spinor *pk,spinor *pl))
{
   (*Op)(psp,psw);
   (*Op)(psw,psap);

   ai=rsq/norm_square(nsp,com,psw);
   update_g();

   rsq_old=rsq;
   rsq=norm_square(nsp,com,psr);
   bi=rsq/rsq_old;
   update_xp();
}


static void accumulate_x(void)
{
   float *rsx,*rsm;
   double *rdx;

   rdx=(double*)(pdx);
   rsm=(float*)(psx+nsp);

   for (rsx=(float*)(psx);rsx<rsm;rsx+=8)
   {
      *(rdx  )+=(double)(*(rsx  ));
      *(rdx+1)+=(double)(*(rsx+1));
      *(rdx+2)+=(double)(*(rsx+2));
      *(rdx+3)+=(double)(*(rsx+3));
      *(rdx+4)+=(double)(*(rsx+4));
      *(rdx+5)+=(double)(*(rsx+5));
      *(rdx+6)+=(double)(*(rsx+6));
      *(rdx+7)+=(double)(*(rsx+7));

      rdx+=8;
   }
}


static void cg_reset(void (*Op)(spinor *pk,spinor *pl),
                     void (*Op_dble)(spinor_dble *pk,spinor_dble *pl))
{
   float r,*rg,*rm;
   double *re,*rv;
   spinor *rsr,*rsp,*rsw,*rsm;
   complex z;

   (*Op_dble)(pdx,pdw);
   (*Op_dble)(pdw,pdv);

   re=(double*)(pdb);
   rv=(double*)(pdv);
   rm=(float*)(psr+nsp);

   for (rg=(float*)(psr);rg<rm;rg+=8)
   {
      *(rg  )=(float)((*(re  ))-(*(rv  )));
      *(rg+1)=(float)((*(re+1))-(*(rv+1)));
      *(rg+2)=(float)((*(re+2))-(*(rv+2)));
      *(rg+3)=(float)((*(re+3))-(*(rv+3)));
      *(rg+4)=(float)((*(re+4))-(*(rv+4)));
      *(rg+5)=(float)((*(re+5))-(*(rv+5)));
      *(rg+6)=(float)((*(re+6))-(*(rv+6)));
      *(rg+7)=(float)((*(re+7))-(*(rv+7)));

      re+=8;
      rv+=8;
   }

   rsq=norm_square(nsp,com,psr);

   rsp=psp;
   rsw=psw;
   rsm=psr+nsp;

   for (rsr=psr;rsr<rsm;rsr++)
   {
      *rsw=*rsp;
      *rsp=*rsr;

      rsp+=1;
      rsw+=1;
   }

   z=spinor_prod(nsp,com,psr,psw);
   z.re=-z.re/rsq;
   z.im=-z.im/rsq;
   mulc_spinor_add(nsp,psw,psr,z);

   (*Op)(psw,psx);
   (*Op)(psx,psap);

   r=norm_square(nsp,com,psx);
   z=spinor_prod(nsp,com,psap,psr);

   if ((z.re*z.re+z.im*z.im)<(2.0f*r*r))
   {
      z.re=-z.re/r;
      z.im=-z.im/r;
      mulc_spinor_add(nsp,psp,psw,z);
   }

   set_s2zero(nsp,psx);
}


double cgne(int vol,int icom,void (*Op)(spinor *pk,spinor *pl),
            void (*Op_dble)(spinor_dble *pk,spinor_dble *pl),
            spinor *ws[],spinor_dble *wsd[],int nmx,double res,
            spinor_dble *pk,spinor_dble *pl,int *status)
{
   int ncg,iprms[2];
   double xn,rn,tol,dprms[1];

   if ((icom==1)&&(NPROC>1))
   {
      iprms[0]=vol;
      iprms[1]=nmx;
      dprms[0]=res;

      MPI_Bcast(iprms,2,MPI_INT,0,MPI_COMM_WORLD);
      MPI_Bcast(dprms,1,MPI_DOUBLE,0,MPI_COMM_WORLD);

      error((iprms[0]!=vol)||(iprms[1]!=nmx)||(dprms[0]!=res),1,
            "cgne [cgne.c]","Parameters are not global");

      error_root((vol<=0)||(nmx<1)||(res<=DBL_EPSILON),1,
                 "cgne [cgne.c]","Improper choice of vol,nmx or res");
   }
   else if (error_loc((vol<=0)||(nmx<1)||(res<=DBL_EPSILON),1,
                      "cgne [cgne.c]","Improper choice of vol,nmx or res"))
   {
      (*status)=0;
      return 1.0;
   }

   cg_init(vol,icom,ws,wsd,pk,pl);

   rn=sqrt((double)(rsq));
   tol=res*rn;
   (*status)=0;

   xn=(double)(norm_square(nsp,com,psx));
   xn=sqrt(xn);

   while (rn>tol)
   {
      /* if (my_rank==0)
         printf("rn_old =%.2e\n",rn); */

      ncg=0;

      for (;;)
      {
         cg_step(Op);
         ncg+=1;
         (*status)+=1;

         xn=(double)(norm_square(nsp,com,psx));
         xn=sqrt(xn);
         rn=sqrt((double)(rsq));

         /*  if ((my_rank==0)&&((ncg%10)==0))
             printf("ncg=%d, xn=%.2e, rn=%.2e\n",(*status),xn,rn); */

         if ((rn<=tol)||(rn<=(PRECISION_LIMIT*xn))||(ncg>=100)||
             ((*status)>=nmx))
            break;
      }

      accumulate_x();
      xn=norm_square_dble(nsp,com,pdx);
      xn=sqrt(xn);
      cg_reset(Op,Op_dble);
      rn=sqrt((double)(rsq));

      if (((*status)>=nmx)&&(rn>tol))
      {
         (*status)=-1;
         break;
      }

      if ((100.0*DBL_EPSILON*xn)>tol)
      {
         (*status)=-2;
         break;
      }
   }

   return rn;
}

