/*******************************************************************************
*
* File blk_sw_term.c
*
* Copyright (C) 2005 Luigi Del Debbio, Martin Luescher
*
* This software is distributed under the terms of the GNU General Public
* License (GPL)
*
* Computation of the SW-term on the blocks of the HMC_BLOCKS grid
*
* The externally accessible functions are
*
*   void swb_term(void)
*     Computes the contributions to the SW-term that do not depend on the
*     active links and stores the calculated matrices in the swb arrays
*     on the blocks of the HMC_BLOCKS grid.
*      This program recalculates the global SW-term if the swd array is not
*     up-to-date and it terminates with an error message if the double-
*     precision gauge fields on the blocks do not match the global field.
*     On exit the swd arrays on the blocks and the global lattice contain
*     the correct SW-term
*
*   void blk_sw_term(int n)
*     Computes the SW-term for the current double-precision gauge field
*     on the n'th block in the HMC_BLOCKS grid and assigns the calculated
*     matrices to the swd array on the block. This program assumes the swb
*     array to be up-to-date. Otherwise no operations are performed and
*     an error message is recorded by the error_loc() function
*
*   void update_swdbgr2swd(void)
*     Computes the double-precision SW-term on all blocks of the HMC_BLOCKS
*     grid (where it is not up-to-date or inverted) and assigns the block
*     swd arrays to the global swd array. This program terminates with an
*     error message if the inactive link variables on the blocks differ 
*     from the global ones or if the swb array is not up-to-date
*
*   void chmass_swdbgr2swd(double kappa)
*     Sets the lattice parameter kappa to the specified value and updates 
*     the swb and swd arrays on the HMC_BLOCKS grid. The block swd arrays
*     are then copied to the global field. This program terminates with an
*     error message if the inactive link variables on the blocks differ 
*     from the global ones or if the swb array was not up-to-date
*
*   void free_blk_swbufs(void)
*     Frees the buffers used for the computation of the SW-term. The program
*     swb_term allocates these buffers automatically if they are not already
*     allocated (in which case they are reused)
*
* Notes:
*
* The normalization and storage conventions for the SW-term are the same as
* in the case of the programs in the module sw_term.c
*
* The parameters m0 and csw are obtained from the parameter data base
* using the programs provided by the module flags/parms.c
*
* The matrices in the swb array depend on the inactive link variables only.
* As long as these are not changed, the swb array thus remains up-to-date.
* In particular, it suffices to call swb_term() at the beginning of each
* HMC trajectory. Along the trajectories the program blk_sw_term() then
* obtains the correct SW-term on the blocks
*
* The programs swb_term() and free_blk_swbufs() perform global operations
* and must be called simultaneously on all processes. The buffers in this
* module are freed automatically when the block grid is freed
*
*******************************************************************************/

#define BLK_SW_TERM_C

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "mpi.h"
#include "su3.h"
#include "flags.h"
#include "misc.h"
#include "start.h"
#include "block.h"
#include "update.h"
#include "sw_term.h"
#include "global.h"

#if (defined SSE2)
#include "sse2.h"
static su3_dble w1,w2 __attribute__ ((aligned (16)));
static const u3_alg_dble qd0 __attribute__ ((aligned (8))) ={0.0};
#else
static su3_dble w1,w2;
static const u3_alg_dble qd0={0.0};
#endif

static int npltot,nplq[6],npofs[6],*plq,*sit;
static u3_alg_dble *Qarr=NULL;


static void u3_alg2pauli1(int vol,u3_alg_dble *X,pauli_dble *m)
{
   u3_alg_dble *Xm;

   Xm=X+vol;

   for (;X<Xm;)
   {
#if (defined SSE2)
      X+=2;
      _prefetch_su3(X);
      X-=2;
      m+=8;
      _prefetch_pauli_dble(m);
      m-=8;
#endif
      (*m).u[10]=-(*X).c1;

      (*m).u[12]=-(*X).c5;
      (*m).u[13]= (*X).c4;
      (*m).u[14]=-(*X).c7;
      (*m).u[15]= (*X).c6;

      (*m).u[18]=-(*X).c5;
      (*m).u[19]=-(*X).c4;
      (*m).u[20]=-(*X).c2;

      (*m).u[22]=-(*X).c9;
      (*m).u[23]= (*X).c8;
      (*m).u[24]=-(*X).c7;
      (*m).u[25]=-(*X).c6;
      (*m).u[26]=-(*X).c9;
      (*m).u[27]=-(*X).c8;
      (*m).u[28]=-(*X).c3;

      *X=qd0;

      m+=2;
      X+=1;
   }
}


static void u3_alg2pauli2(int vol,u3_alg_dble *X,pauli_dble *m)
{
   u3_alg_dble *Xm;

   Xm=X+vol;

   for (;X<Xm;)
   {
#if (defined SSE2)
      X+=2;
      _prefetch_su3(X);
      X-=2;
      m+=8;
      _prefetch_pauli_dble(m);
      m-=8;
#endif   
      (*m).u[11] =(*X).c1;
      (*m).u[12]+=(*X).c4;
      (*m).u[13]+=(*X).c5;
      (*m).u[14]+=(*X).c6;
      (*m).u[15]+=(*X).c7;

      (*m).u[18]-=(*X).c4;
      (*m).u[19]+=(*X).c5;

      (*m).u[21] =(*X).c2;
      (*m).u[22]+=(*X).c8;
      (*m).u[23]+=(*X).c9;
      (*m).u[24]-=(*X).c6;
      (*m).u[25]+=(*X).c7;
      (*m).u[26]-=(*X).c8;
      (*m).u[27]+=(*X).c9;

      (*m).u[29] =(*X).c3;

      *X=qd0;

      m+=2;
      X+=1;      
   }
}


static void u3_alg2pauli3(int vol,u3_alg_dble *X,pauli_dble *m)
{
   u3_alg_dble *Xm;

   Xm=X+vol;

   for (;X<Xm;)
   {
#if (defined SSE2)
      X+=2;
      _prefetch_su3(X);
      X-=2;
      m+=8;
      _prefetch_pauli_dble(m);
      m-=8;
#endif   
      (*m).u[ 0]=-(*X).c1;
      (*m).u[ 1]=-(*X).c2;
      (*m).u[ 2]=-(*X).c3;
      (*m).u[ 3]= (*X).c1;
      (*m).u[ 4]= (*X).c2;
      (*m).u[ 5]= (*X).c3;
      (*m).u[ 6]=-(*X).c5;
      (*m).u[ 7]= (*X).c4;
      (*m).u[ 8]=-(*X).c7;
      (*m).u[ 9]= (*X).c6;

      (*m).u[16]=-(*X).c9;
      (*m).u[17]= (*X).c8;

      (*m).u[30]= (*X).c5;
      (*m).u[31]=-(*X).c4;
      (*m).u[32]= (*X).c7;
      (*m).u[33]=-(*X).c6;
      (*m).u[34]= (*X).c9;
      (*m).u[35]=-(*X).c8;

      *X=qd0;

      m+=2;
      X+=1;      
   }
}


static int is_on_bnd(block_t *b,int ix)
{
   int mu;

   if (ix>=(*b).vol)
      return 1;
   
   for (mu=0;mu<4;mu++)
      if (((*b).iup[ix][mu]>=(*b).vol)||((*b).idn[ix][mu]>=(*b).vol))
         return 1;

   return 0;
}


static void set_plq(void)
{
   int id,mu,nu,nb,isw;
   int ix,iy,iw,iz;
   int ibx,iby,ibw,ibz,*p,*s;
   int plns[6][2]={{0,1},{0,2},{0,3},{2,3},{3,1},{1,2}};
   block_t *b;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   p=plq;
   s=sit;

   for (id=0;id<6;id++)
   {
      mu=plns[id][0];
      nu=plns[id][1];
      
      for (ix=0;ix<(*b).vol;ix++)
      {
         iy=(*b).iup[ix][mu];
         
         if (iy<(*b).vol)
         {
            iw=(*b).iup[ix][nu];
            iz=(*b).iup[iy][nu];
            
            ibx=is_on_bnd(b,ix);
            iby=is_on_bnd(b,iy);
            ibz=is_on_bnd(b,iz);
            ibw=is_on_bnd(b,iw);
            
            if ((ibx+iby+ibz+ibw)!=4)
            {
               if (ix<(*b).volh)
               {
                  p[0]=8*(iy-(*b).volh)+2*mu+1;
                  p[1]=8*(iy-(*b).volh)+2*nu;
                  p[2]=8*(iw-(*b).volh)+2*nu+1;
                  p[3]=8*(iw-(*b).volh)+2*mu;
               }
               else
               {
                  p[0]=8*(ix-(*b).volh)+2*mu;
                  p[1]=8*(iz-(*b).volh)+2*nu+1; 
                  p[2]=8*(ix-(*b).volh)+2*nu;
                  p[3]=8*(iz-(*b).volh)+2*mu+1;;
               }

               s[0]=ix;
               s[1]=iy;
               s[2]=iw;
               s[3]=iz;

               p+=4;               
               s+=4;
            }
         }
      }
   }

   p[0]=plq[0];
   p[1]=plq[1];
   p[2]=plq[2];
   p[3]=plq[3];

   s[0]=sit[0];
   s[1]=sit[1];
   s[2]=sit[2];
   s[3]=sit[3];   
}


static void alloc_blk_swbufs(void)
{
   int *bs,nb,isw;
   u3_alg_dble *qd,*qm;
   block_t *b;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   error(b==NULL,1,"alloc_blk_swbufs [blk_sw_term.c]",
         "Block grid is not allocated"); 

   bs=(*b).bs;
   nplq[0]=(bs[0]-1)*(bs[1]-1)*(bs[2]-2)*(bs[3]-2);
   nplq[1]=(bs[0]-1)*(bs[1]-2)*(bs[2]-1)*(bs[3]-2);
   nplq[2]=(bs[0]-1)*(bs[1]-2)*(bs[2]-2)*(bs[3]-1);
   nplq[3]=(bs[0]-2)*(bs[1]-2)*(bs[2]-1)*(bs[3]-1);
   nplq[4]=(bs[0]-2)*(bs[1]-1)*(bs[2]-2)*(bs[3]-1);
   nplq[5]=(bs[0]-2)*(bs[1]-1)*(bs[2]-1)*(bs[3]-2);

   npofs[0]=0;
   npofs[1]=npofs[0]+nplq[0];
   npofs[2]=npofs[1]+nplq[1];
   npofs[3]=npofs[2]+nplq[2];
   npofs[4]=npofs[3]+nplq[3];
   npofs[5]=npofs[4]+nplq[4];

   npltot=nplq[0]+nplq[1]+nplq[2]+nplq[3]+nplq[4]+nplq[5];
   plq=amalloc(8*(npltot+1)*sizeof(int),3);
   Qarr=amalloc((*b).vol*sizeof(u3_alg_dble),ALIGN);
   error((plq==NULL)||(Qarr==NULL),1,"alloc_blk_swbufs [blk_sw_term.c]",
         "Unable to allocate field arrays");

   sit=plq+4*(npltot+1);
   set_plq();
   qm=Qarr+(*b).vol;

   for (qd=Qarr;qd<qm;qd++)
      *qd=qd0;
}


static void build_Q(block_t *b,int id)
{
   int *pl,*plm,*si;
   su3_dble *ub,*pu0,*pu1,*pu2,*pu3;
   u3_alg_dble *Q0,*Q1;

   ub=(*b).ud[0];
   pl=plq+4*npofs[id];
   plm=pl+4*nplq[id];
   si=sit+4*npofs[id];

   pu0=ub+pl[0];
   pu1=ub+pl[1];
   pu2=ub+pl[2];
   pu3=ub+pl[3];
   
   for (;pl<plm;)
   {
      Q0=Qarr+si[0];
      Q1=Qarr+si[3];

#if (defined SSE2)
      _prefetch_su3(Q0);
      _prefetch_su3(Q1);      
#endif      

      su3xsu3(pu0,pu1,&w1);
      su3dagxsu3dag(pu3,pu2,&w2);
      add_prod2u3alg(&w1,&w2,Q0);
      add_prod2u3alg(&w2,&w1,Q1);

      Q0=Qarr+si[1];
      Q1=Qarr+si[2];
      si+=4;

#if (defined SSE2)
      _prefetch_su3(Q0);
      _prefetch_su3(Q1);      
#endif 
      
      su3dagxsu3(pu2,pu0,&w1);

      pl+=4;
      pu0=ub+pl[0];
      pu2=ub+pl[2];

#if (defined SSE2)
      _prefetch_su3_dble(pu0);
      _prefetch_su3_dble(pu2);      
#endif      

      su3xsu3dag(pu1,pu3,&w2);

      pu1=ub+pl[1];
      pu3=ub+pl[3];

#if (defined SSE2)
      _prefetch_su3_dble(pu1);
      _prefetch_su3_dble(pu3);      
#endif      

      add_prod2u3alg(&w2,&w1,Q0);              
      add_prod2u3alg(&w1,&w2,Q1);
   }
}


static void blk_build_F(block_t *b)
{
   int bvol=(*b).vol;
   
   build_Q(b,0);
   u3_alg2pauli1(bvol,Qarr,(*b).swd);

   build_Q(b,1);
   u3_alg2pauli2(bvol,Qarr,(*b).swd);

   build_Q(b,2);
   u3_alg2pauli3(bvol,Qarr,(*b).swd);

   build_Q(b,3);
   u3_alg2pauli1(bvol,Qarr,(*b).swd+1);

   build_Q(b,4);
   u3_alg2pauli2(bvol,Qarr,(*b).swd+1);

   build_Q(b,5);
   u3_alg2pauli3(bvol,Qarr,(*b).swd+1);
}


static void active_blk_sw(block_t *b)
{
   double c2;
   double *rup,*rdn,*rm,r1,r2;
   pauli_dble *pd,*pm;
   lat_parms_t lat;
   
   blk_build_F(b);

   lat=lat_parms();   
   c2=lat.csw/16.0;

   pd=(*b).swd;
   pm=pd+2*(*b).vol;

   for (;pd<pm;pd+=2)
   {
      rup=(*(pd  )).u;
      rdn=(*(pd+1)).u;
      rm=rup+36;

      for (;rup<rm;rup+=1)
      {
         r1=*rup;
         r2=*rdn;
         *rup=c2*(r1-r2);
         *rdn=c2*(r1+r2);
         rdn+=1;
      }
   }
}


static void blk_swb_term(block_t *b)
{
   double *ru,*rm,*rb;
   pauli_dble *pd,*pm,*pb;
   
   pd=(*b).swd;
   (*b).swd=(*b).swb;
   
   active_blk_sw(b);

   (*b).swd=pd;
   pm=pd+2*(*b).vol;
   pb=(*b).swb;

   for (;pd<pm;pd++)
   {
      ru=(*pd).u;
      rb=(*pb).u;
      rm=ru+36;
      
      for (;ru<rm;ru++)
      {
         *rb=(*ru)-(*rb);
         rb+=1;
      }
      
      pb+=1;
   }
}


void swb_term(void)
{
   int nb,isw;
   block_t *b,*bm;

   if (Qarr==NULL)
      alloc_blk_swbufs();
   
   b=blk_list(HMC_BLOCKS,&nb,&isw);

   error(query_blk_flags(HMC_BLOCKS,UDBLK_MATCH_UD,nb)!=1,1,
         "swb_term [blk_sw_term.c]",
         "Block gauge fields do not match the global field");
   
   if ((query_flags(SWD_UP2DATE)!=1)||
       (query_flags(SWD_E_INVERTED)!=0)||(query_flags(SWD_O_INVERTED)!=0))
      sw_term();

   assign_swd2swdbgr(HMC_BLOCKS,NO_PTS);

   error(check_blk_flags(HMC_BLOCKS,SW_TERM_INACTIVE,nb)!=0,1,
         "swb_term [blk_sw_term.c]","Fields are not in proper condition");

   bm=b+nb;
   
   for (;b<bm;b++)
      blk_swb_term(b);

   set_blk_flags(HMC_BLOCKS,COMPUTED_SWB,nb);
}


void blk_sw_term(int n)
{
   int nb,isw;
   double *ru,*rm,*rb;
   pauli_dble *pd,*pm,*pb;
   block_t *b;

   if (error_loc((Qarr==NULL)||(check_blk_flags(HMC_BLOCKS,SW_TERM,n)!=0),1,
                 "blk_sw_term [blk_sw_term.c]",
                 "Fields are not in proper condition"))
      return;
   
   b=blk_list(HMC_BLOCKS,&nb,&isw)+n;

   active_blk_sw(b);

   pd=(*b).swd;
   pm=pd+2*(*b).vol;
   pb=(*b).swb;

   for (;pd<pm;pd++)
   {
      ru=(*pd).u;
      rb=(*pb).u;
      rm=ru+36;
      
      for (;ru<rm;ru++)
      {
         *ru+=*rb;
         rb+=1;
      }

      pb+=1;
   }
   
   set_blk_flags(HMC_BLOCKS,COMPUTED_SWD,n);
}


static void assign_swdbgr2swd(void)
{
   int nb,isw,*imb;
   pauli_dble *m,*mm,*ms;
   block_t *b,*bm;

   b=blk_list(HMC_BLOCKS,&nb,&isw);
   bm=b+nb;

   for (;b<bm;b++)
   {
      m=(*b).swd;
      mm=m+2*(*b).vol;
      imb=(*b).imb;

      for (;m<mm;m+=2)
      {
         ms=swd+2*(*imb);
         imb+=1;

         *ms=*m;
         *(ms+1)=*(m+1);
      }
   }

   set_blk_flags(HMC_BLOCKS,ASSIGNED_SWDBGR2SWD,nb);
}


void update_swdbgr2swd(void)
{
   int n,nb,isw;

   blk_list(HMC_BLOCKS,&nb,&isw);
   
   error(check_blk_flags(HMC_BLOCKS,UPDATE_SWDBGR2SWD,nb)!=0,1,
         "update_swdbgr2swd [blk_sw_term.c]",
         "Fields are not in proper condition");

   for (n=0;n<nb;n++)
   {
      if ((query_blk_flags(HMC_BLOCKS,SWD_UP2DATE,n)!=1)||
          (query_blk_flags(HMC_BLOCKS,SWD_E_INVERTED,n)!=0)||
          (query_blk_flags(HMC_BLOCKS,SWD_O_INVERTED,n)!=0))
         blk_sw_term(n);      
   }

   assign_swdbgr2swd();   
}


static void shift_swd(int vol,double dm,pauli_dble *m)
{
   pauli_dble *mm;
   
   mm=m+2*vol;

   for (;m<mm;)
   {
      (*m).u[0]+=dm;
      (*m).u[1]+=dm;
      (*m).u[2]+=dm;
      (*m).u[3]+=dm;
      (*m).u[4]+=dm;
      (*m).u[5]+=dm;         
      m+=1;

      (*m).u[0]-=dm;
      (*m).u[1]-=dm;
      (*m).u[2]-=dm;
      (*m).u[3]-=dm;
      (*m).u[4]-=dm;
      (*m).u[5]-=dm;         
      m+=1;
   }
}


void chmass_swdbgr2swd(double kappa)
{
   int n,nb,isw;
   double m0,dm0;
   block_t *b,*bm;
   lat_parms_t lat;

   b=blk_list(HMC_BLOCKS,&nb,&isw);

   error(check_blk_flags(HMC_BLOCKS,CHMASS_SWDBGR2SWD,nb)!=0,1,
         "chmass_swdbgr2swd [blk_sw_term.c]",
         "Fields are not in proper condition");

   for (n=0;n<nb;n++)
   {
      if ((query_blk_flags(HMC_BLOCKS,SWD_UP2DATE,n)!=1)||
          (query_blk_flags(HMC_BLOCKS,SWD_E_INVERTED,n)!=0)||
          (query_blk_flags(HMC_BLOCKS,SWD_O_INVERTED,n)!=0))
         blk_sw_term(n);      
   }
   
   lat=lat_parms();
   m0=lat.m0;
   lat=set_lat_parms(lat.beta,kappa,lat.csw);
   dm0=lat.m0-m0;
   bm=b+nb;

   for (;b<bm;b++)
   {
      shift_swd((*b).vol,dm0,(*b).swb);
      shift_swd((*b).vol,dm0,(*b).swd);
   }

   set_blk_flags(HMC_BLOCKS,COMPUTED_SWB,nb);   
   set_blk_flags(HMC_BLOCKS,COMPUTED_SWD,nb);
   
   assign_swdbgr2swd();
}


void free_blk_swbufs(void)
{
   if (Qarr==NULL)
      return;

   afree(Qarr);
   afree(plq);
   Qarr=NULL;
}
